/*  File:     Methods.pl
    Purpose:  Basic reasoning methods for GARP
    Author:   Martin Reinders & Bert Bredeweg & Floris Linnebank
    Date:     August 1989
    Part-of:  GARP (version 2.0)
    Modified: 30 July 2004

    Copyright (c) 2004, University of Amsterdam. All rights reserved.
    
*/

% new input model

new_input_model(Name):-
    clean_up, 
    initialise_quantity_spaces, 
    SMD =.. [ smd, Name, SE, P, V, R, SS ], 
    call(SMD), 
    % nest input smd (augmented with dummy argument) in overall smd
    ISMD =.. [smd, Name, SE, P, V, R, SS, nil ], 
    smd_slots(NSMD, Name, [], [], [], [], [], ISMD), 
    Name = input_system( RealName ),
    etrace(methods_start_sim, [RealName], general),
    successor_states(
        to(cause([]), conditions([]), results([]), to_state([]), _), 
        input, NSMD, _).
/**/

/*
 * abolish global predicaties state & qspace
 * reset gensym
 *
 */
 
clean_up:-
    %once((reset_gensym ; true)),  % Hack: reset_gensym seems evil, does commenting it have any consequences?
    retractall(state(_, _)), 
    retractall(state_status(_, _)), 
    retractall(state_to(_, _)), 
    retractall(state_from(_, _)), 
    retractall(qspace(_, _, _, _)), 
    flag(state_number, _, 0).

make_new_state(Name, Status, FromStates, ToList, SMD):-
    flag(state_number, Old, Old), 
    Name is Old + 1, 
    flag(state_number, _, Name), 
    assertz(state(Name, SMD) ), 
    assertz(state_status(Name, Status)), 
    assertz(state_from(Name, FromStates)), 
    assertz(state_to(Name, ToList)), 
    !.


/* --------------------------- TERMINATION --------------------------- */

terminate_all :-
    forall(state_status(State, interpreted), termination(State) ).


%*** FL 10-3-2004: Old rule based version ACTIVE UNDER SWITCH ***

% find terminations for a state
    
termination(State):-
    flag(use_old_transition_rules, Flag, Flag), % New
    algorithm_assumption_flag(Flag, true),      % New

    state(State, SMD), 
    retract(state_status(State, interpreted)), 
    retract(state_to(State, [])), 

        % Have GARP produce XML comments when running via the internet...
        (  current_prolog_flag( bbwww_garp, true )
        -> xml_comment( 2, ['determining terminations for state ', State], garpinfo)
        ;  etrace(methods_det_term, [State], general) 
        ),
 
    findall(
        to( cause(Causes), 
            conditions(Cond), 
            results(Results), 
            to_state([]), 
            terminated), 
        (   rule(termination, Causes, condition(Cond), 
                        result(Results) ), 
            check_rule_conditions(Cond, SMD, [])
        ), 
        To), 
    assert(state_status(State, terminated)), 
    assert(state_to(State, To)),
    etrace(methods_show_terminations, [State, To], termination),
    !.

%*** GARP 2.0 Version below using procedures & inequality reasoning: ***

% find terminations for a state

termination(State):-
    flag(use_old_transition_rules, Flag, Flag),
    algorithm_assumption_flag(Flag, fail),
    state(State, SMD), 
    retract(state_status(State, interpreted)), 
    retract(state_to(State, [])), 

        % Have GARP produce XML comments when running via the internet...
        (  current_prolog_flag( bbwww_garp, true )
        -> xml_comment( 2, ['determining terminations for state ', State], garpinfo)
        ;  etrace(methods_det_term, [State], general)
        ),
    
    findall(To, single_termination(SMD, To), ToList1), 

    % extra rule interpreter for domain specific terminations
    flag(extra_termination_interpreter, ETIFlag, ETIFlag),
    (
        algorithm_assumption_flag(ETIFlag, true)
        ->
        findall(to( cause(Causes), 
                conditions(Cond), 
                results(Results), 
                to_state([]), 
                terminated), 
                (
                rule(termination, Causes, condition(Cond), 
                     result(Results) ), 
                check_rule_conditions(Cond, SMD, [])
                ), 
                ToList2) 
        ;
        ToList2 = []
    ),
    
    append(ToList1, ToList2, ToList),
    assert(state_status(State, terminated)), 
    assert(state_to(State, ToList)),
    etrace(methods_show_terminations, [State, ToList], termination),
    !.

% END NEW


/* ------------------------ PRECEDENCE ----------------------------- */

order_all :-
    forall(state_status(State, terminated), precedence(State) ).


%*** FL March 2004: old rule based precedence: ACTIVE UNDER SWITCH***

precedence(State):-
    flag(use_old_transition_rules, Flag, Flag), % New
    algorithm_assumption_flag(Flag, true),      % New
    state(State, SMD), 
    retract(state_status(State, terminated) ), 
    retract(state_to(State, Terminated) ), 

        % Have GARP produce XML comments when running via the internet...
        (  current_prolog_flag( bbwww_garp, true )
        -> xml_comment( 2, ['determining ordering for state ', State], garpinfo)
        ;  etrace(methods_det_ord, [State], ordering)
        ),

    apply_order_rules(Terminated, Ordered, SMD), 
    merge_possible_terminations(Ordered, ToFinal1), 
    set_to_list_status_ordered(ToFinal1, ToFinal), 
    asserta(state_status(State, ordered) ), 
    asserta(state_to(State, ToFinal) ), 
    etrace(methods_show_ord, [State, ToFinal], ordering),
    !.

%*** New garp 2.0 version with procedures & inequality reasoning below: ***

precedence(State):-
    flag(use_old_transition_rules, Flag, Flag),
    algorithm_assumption_flag(Flag, fail),
    state(State, SMD), 
    retract(state_status(State, terminated) ), 
    retract(state_to(State, Terminated) ), 

        % Have GARP produce XML comments when running via the internet...
        (  current_prolog_flag( bbwww_garp, true )
        -> xml_comment( 2, ['determining ordering for state ', State], garpinfo)
        ;  etrace(methods_det_ord, [State], general)
        ),

    do_precedence(SMD, Terminated, ToFinal),

    asserta(state_status(State, ordered) ), 
    asserta(state_to(State, ToFinal) ),
    etrace(methods_show_ord, [State, ToFinal], ordering),
    !.
    
% End New


% when all order rules have been applied set status of each termination
% to ordered.

set_to_list_status_ordered([], []).
set_to_list_status_ordered([to(Cause, Cond, Result, ToStates, terminated)|T1], 
           [to(Cause, Cond, Result, ToStates, ordered)|T2]) :-
set_to_list_status_ordered(T1, T2).

apply_order_rules(To1, To2, SMD):-
    rule(precedence, Cause, condition(Conditions), action(Actions)), 
    check_rule_conditions(Conditions, SMD, To1), 
    tracer(ordering, 'apply: %w', [Cause]), 
    apply_order_actions(Cause, Actions, To1, NTo1), 
    !, 
    apply_order_rules(NTo1, To2, SMD).
apply_order_rules(To, To, _).


apply_order_actions(_, [], To, To).

apply_order_actions(Cause, [ remove(L) | T ], To, NTo):-
    remove_from_tolist(L, To, TTo),     % what to do with Cause?
    (@tracer->>tell(ordering) -> %gp3 0.3
        write_ln('removing: '), 
        pr_l_tab(4, L),
        @tracer->>told;
        true
    ), 
    apply_order_actions(Cause, T, TTo, NTo).

apply_order_actions(Cause, [ merge(L) | T ], To, NTo):-
    merge_tolist_select(Cause, L, To, TTo, New), 
    (trace_on(ordering) ->
        write('merging :'), 
        pr_l_tab(4, L)
        ;
        true
    ), 
    apply_order_actions(Cause, T, [New|TTo], NTo).

remove_from_tolist([], To, To).
remove_from_tolist([H|T], To, NTo):-
    common_select(To, to(cause(Causes), _, _, _, _Status), TTo), 
    member(H, Causes), 
    !, 
    remove_from_tolist(T, TTo, NTo).

% merge a termination 'Cause' with another termination

merge_tolist_select(Cause, [], To, To, 
    to(cause([Cause]), conditions([]), results([]),     to_state([]), _Status) ):- !.

merge_tolist_select(Cause, [H|T], To, NTo, 
        to( cause(Causes), 
            conditions(Conditions), 
            results(Results), 
            to_state(States), Status) ) :-
    merge_tolist_select(Cause, T, To, TTo, 
        to( cause(Causes2), 
            conditions(Conditions2), 
            results(Results2), 
            to_state(States2), Status)), 
    common_select(TTo, 
        to( cause(Causes1), 
            conditions(Conditions1), 
            results(Results1), 
            to_state(States1), Status), 
        NTo), 
    memberchk(H, Causes1), 
    smerge(Causes1, Causes2, Causes), 
    merge_rule_conditions_givens(Conditions1, Conditions2, Conditions), 
    merge_rule_conditions_givens(Results1, Results2, Results), 
    smerge(States1, States2, States), 
    !.

% merge a list of terminations

merge_tolist([], to(cause([]), 
            conditions([]), 
            results([]), 
            to_state([]), _Status) ):- !.

merge_tolist([
    to( cause(Causes1), 
        conditions(Conditions1), 
        results(Results1), 
        to_state(States1), Status)
    |Tail], 
    to( cause(Causes), 
        conditions(Conditions), 
        results(Results), 
        to_state(States), Status) )    :-
    merge_tolist(Tail, 
    to( cause(Causes2), 
        conditions(Conditions2), 
        results(Results2), 
        to_state(States2), Status)), 
    smerge(Causes1, Causes2, Causes), 
    merge_rule_conditions_givens(Conditions1, Conditions2, Conditions), 
    merge_rule_conditions_givens(Results1, Results2, Results), 
    smerge(States1, States2, States), 
    !.


merge_possible_terminations(ToList, Final):-
    findall(To, 
        (   a_subset(Subset, ToList), 
            merge_tolist(Subset, To)
        ), 
        Final).

/* ----------------------- (UNDO) TRANSITION --------------------------- */

close_all:-
    forall(state_status(State, ordered), transition(State) ).

% no terminations for this state

transition(State):-
    state_to(State, []), 
    retract(state_status(State, ordered) ), 
    assert(state_status(State, closed) ), 
    !.

transition(State):-
    transition(State, _), 
    fail.

transition(_).

transition(State, Which):-
    state(State, SMD), 
    state_to(State, ToList), 
    !, 
    nth1(Which, ToList, to(Cause, Cond, Result, _, ordered)), 
    
        % Have GARP produce XML comments when running via the internet...
        (  current_prolog_flag( bbwww_garp, true )
        -> xml_comment( 2, ['working on transition ', Which, ' of state ', State], garpinfo)
        ;  etrace(methods_do_trans, [Which, State], general) 
        ),

    successor_states(to(Cause, Cond, Result, _, ordered), State, SMD, NewTo), 
    once((  retract(state_to(State, ToOld  )), 
        replace_nth1(Which, ToOld, NewTo, ToNew), 
        asserta(state_to(State, ToNew) )
         )), 
    (memberchk(to(_, _, _, _, ordered), ToNew) ->
        true
        ;
        retract(state_status(State, _) ), 
        assert(state_status(State, closed) )
    ).


undo_transition(State):-
    state_to(State, []), 
    retract(state_status(State, closed) ), 
    asserta(state_status(State, ordered) ), 
    !.

undo_transition(State):-
    undo_transition(State, _), 
    fail.

undo_transition(_).

undo_transition(State, Which):-
    state_to(State, ToList), 
    !, 
    nth1(Which, ToList, to(Cause, Cond, Result, to_state(ToStates), closed)), 
    writef('undo transition %w of state %w to states:\n', [Which, State]), 
    pr_l_tab(4, ToStates), 
    once((  retract(state_to(State, OldTo) ), 
        replace_nth1(Which, OldTo, 
            to(Cause, Cond, Result, to_state([]), ordered), NewTo), 
        asserta(state_to(State, NewTo) )
         )), 
    undo_successor_states(ToStates, State).


successor_states(to(Cause, Cond, Res, to_state([]), ordered), From, SMD, 
              to(Cause, Cond, Res, to_state(States), closed)):-
    findall(State, depth(to(Cause, Cond, Res, to_state([]), ordered), SMD, State), 
        States), 
    forall(member(St, States), append_from(St, From)), 
    !.


undo_successor_states([], _).
undo_successor_states([H|T], From):-
    state_from(H, FromList), 
    common_select(FromList, From, []), % only predecessor: retract state
    !, 
    undo_transition(H),         % and all its successors
    retract(state(H, _)), 
    retract(state_status(H, _)), 
    retract(state_to(H, _)), 
    retract(state_from(H, _)), 
    undo_successor_states(T, From).

undo_successor_states([H|T], From):-
    state_from(H, FromList), 
    common_select(FromList, From, NewFromList),     % more predecessors
    retract(state_from(H, _)), 
    asserta(state_from(H, NewFromList)), 
    !, 
    undo_successor_states(T, From).

% merge state From into from list of successor State

append_from(State, From):-
    retract(state_from(State, Fr)), 
    smerge([From], Fr, NFr), 
    asserta(state_from(State, NFr)), 
    !.

/* -------------------------- remove from smd ----------------------- */

remove_from_smd([], ResultSMD, ResultSMD).

remove_from_smd([ parameters(PList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    remove_all(PList, P, NP), 
    smd_slots(SMD2, Name, SE, NP, V, R, SS, IS), 
    remove_from_smd(Tail, SMD2, ResultSMD).

remove_from_smd([ par_values(VList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    remove_all(VList, V, NV), 
    smd_slots(SMD2, Name, SE, P, NV, R, SS, IS), 
    remove_from_smd(Tail, SMD2, ResultSMD).
    
remove_from_smd([ par_relations(RList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    % get alternative notations for RList, e.g. A > B = B < A
    findall(AR, 
        (member(Re, RList), alternative_relation(Re, AR) ), ARS), 
    append(ARS, RList, ARList), 
    remove_all(ARList, R, NR), 
    smd_slots(SMD2, Name, SE, P, V, NR, SS, IS), 
    remove_from_smd(Tail, SMD2, ResultSMD).

% system elements are special: if we want to remove
% intance(water1, substance) (a condition of a termination rule)
% then we actually have to remove each instance relation that
% mentions water1 (the most likely is instance(water1, water)).

remove_from_smd([ system_elements(SEList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    remove_system_elements(SEList, SE, NSE), 
    smd_slots(SMD2, Name, NSE, P, V, R, SS, IS), 
    remove_from_smd(Tail, SMD2, ResultSMD).

remove_from_smd([ system_structures(SSList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    remove_all(SSList, SS, NSS), 
    smd_slots(SMD2, Name, SE, P, V, R, NSS, IS), 
    remove_from_smd(Tail, SMD2, ResultSMD).

% allow action tests etc.

remove_from_smd([ _ | Tail ], SMD, ResultSMD):-
    remove_from_smd(Tail, SMD, ResultSMD).

/* --------------------- add to smd ------------------------------- */

add_to_smd([], ResultSMD, ResultSMD).

add_to_smd([ parameters(PList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    smerge(PList, P, NP), 
    smd_slots(SMD2, Name, SE, NP, V, R, SS, IS), 
    add_to_smd(Tail, SMD2, ResultSMD).

add_to_smd([ par_values(VList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    merge_values(VList, V, NV), 
    smd_slots(SMD2, Name, SE, P, NV, R, SS, IS), 
    add_to_smd(Tail, SMD2, ResultSMD).
    
add_to_smd([ par_relations(RList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    smerge(RList, R, NR), 
    smd_slots(SMD2, Name, SE, P, V, NR, SS, IS), 
    add_to_smd(Tail, SMD2, ResultSMD).

add_to_smd([ system_elements(EList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    smerge(EList, SE, NSE), 
    smd_slots(SMD2, Name, NSE, P, V, R, SS, IS), 
    add_to_smd(Tail, SMD2, ResultSMD).

add_to_smd([ system_structures(SList) | Tail ], SMD1, ResultSMD):-
    !, 
    smd_slots(SMD1, Name, SE, P, V, R, SS, IS), 
    smerge(SList, SS, NSS), 
    smd_slots(SMD2, Name, SE, P, V, R, NSS, IS), 
    add_to_smd(Tail, SMD2, ResultSMD).

% allow action tests etc.

add_to_smd([ _ | Tail ], SMD, ResultSMD):-
    add_to_smd(Tail, SMD, ResultSMD).


/* -------------------- merge rule conditions/results --------------- */

% merge two sets of results or conditions

% values must match, like add to smd ..

merge_rule_conditions_givens([ par_values(V1) | Tail ], L2, Result):-
    common_select(L2, par_values(V2), RL2), 
    !, 
    merge_values(V1, V2, NV), 
    merge_rule_conditions_givens(Tail, [par_values(NV)|RL2], Result).

% anything else

% functor(List) in both sets

merge_rule_conditions_givens([S1|T1], L2, R):-
    S1 =.. [Functor, P1 ], 
    common_select(L2, S2, T2), 
    S2 =.. [Functor, P2], 
    !, 
    smerge(P1, P2, P3), 
    S3 =.. [Functor, P3], 
    merge_rule_conditions_givens(T1, [S3|T2], R).

% functor(List) in first set only

merge_rule_conditions_givens([S1|T1], L2, R):-
    merge_rule_conditions_givens(T1, [S1|L2], R).

merge_rule_conditions_givens([], R, R) :- !.

/* ----------------- apply transition rules ---------------------- */

% at the moment we only apply transition rules on values: one rule for
% each value

apply_transition_rules(values, [], [], []).
apply_transition_rules(values, [OnFirst|Rest], Conditions, Givens):-
    rule(transition, _, condition(ThisConditions), givens(ThisGivens)), 
    ThisConditions = [ par_values([ OnFirst ]) ], 
        % not really very much to chose from
    apply_transition_rules(values, Rest, ConditionsRest, GivensRest), 
    append(ThisGivens, GivensRest, Givens), 
    append(ThisConditions, ConditionsRest, Conditions).


remove_all([], L, L).

remove_all([H|T], L, NL):-
    common_select(L, H, L1), 
    !, 
    remove_all(T, L1, NL).

remove_all([_|T], L, NL):-
    remove_all(T, L, NL).

% check rule conditions for termination rules

check_rule_conditions([], _, _):- !.

check_rule_conditions([ termination(TList) | Tail ], SMD, To):-
    !, 
    check_termination_conditions(TList, To), 
    check_rule_conditions(Tail, SMD, To).

check_rule_conditions([ system_elements(EList) | Tail ], SMD, To):-
    !, 
    smd_slots(SMD, _, Elements, _, _, _, _, _), 
    super_class_sub_set(EList, Elements), 
    % Elist specifies superconcept relations on a sub set of Elements
    check_rule_conditions(Tail, SMD, To).

check_rule_conditions([ parameters(PList) | Tail ], SMD, To):-
    !, 
    smd_slots(SMD, _, _, Parameters, _, _, _, _), 
    sub_set(PList, Parameters), 
    check_rule_conditions(Tail, SMD, To).

check_rule_conditions([ par_values(VList) | Tail ], SMD, To):-
    !, 
    smd_slots(SMD, _, _, _, Values, _, _, _), 
    value_sub_set(VList, Values), 
    check_rule_conditions(Tail, SMD,  To).

check_rule_conditions([ par_relations(RList) | Tail ], SMD, To):-
    !, 
    smd_slots(SMD, _, _, _, _, Relations, _, _), 

        % REPAIR: 23/08/2001 by BB (adding VList)
    smd_slots(SMD, _, _, _, VList, _, _, _), 
    relations_sub_set(RList, Relations, VList), 

    check_rule_conditions(Tail, SMD, To).

check_rule_conditions([ quantity_spaces(QList) | Tail ], SMD, To):-
    !, 
    check_quantity_spaces(QList), 
    check_rule_conditions(Tail, SMD, To).

check_rule_conditions([ H | _ ], _, _):-
    writef('*** Error: can not apply rule condition %w\n', [H]), 
    fail.


% condition specifies a termination rule

% REPAIR: 18/04/2001 by BB
% The old version did not try all possibilities. If the 2nd
% clause failed, it removed the 1st 'to' from the 'Termination' list.
% After that, those removed terminations are lost, although they
% may have been crucial for one of the conditions still present
% in the 'Tail' of the '[H|T]' conditions of the rule that is
% being evaluated.
%
% The new version 'removes' only the item from the 'Termination'
% list that matches the 'Header' of the '[H|T]' conditions of the
% rule that is being evaluated (using 'select', a system predicate).
% All the other (remaining) terminations are used to evaluate the
% next condition in the 'Tail' of the '[H|T]' of the rule.
%
% OLD:
% check_termination_conditions([], _) :- !.
% check_termination_conditions([H|T], [to(cause(Causes), _, _, _, _)|R]):-
%   member(H, Causes), 
%   check_termination_conditions(T, R).
% check_termination_conditions(L, [_|R]):-
%   check_termination_conditions(L, R).
%
% NEW:
check_termination_conditions([], _) :- !.
check_termination_conditions( [H|T], Terminations ) :-
        select( to(cause(Causes), _, _, _, _), Terminations, Rest ),
        member( H, Causes ),
        check_termination_conditions( T, Rest ).


/* rewrite quantityspace list so that each interval i becomes interval(i) */

add_interval_structure([], []).

add_interval_structure([point(X)|T], [point(X)|NT]):-
    !, 
    add_interval_structure(T, NT).

add_interval_structure([I|T], [interval(I)|NT]):-
    add_interval_structure(T, NT).

% quantity space conditions

check_quantity_spaces([]).
check_quantity_spaces([meets(Par, L)|T]):-
    !, 
    qspace(Par, _, Space, _), 
    add_interval_structure(Space, NSpace), 
    part_list(L, NSpace), 
    check_quantity_spaces(T).


check_quantity_spaces([interval(Par, I)|T]):-
    !, 
    qspace(Par, _, Space, _), 
    memberchk(I, Space), 
    I \= point(_),          % shouldn't be possible anyway
    check_quantity_spaces(T).
check_quantity_spaces([point(Par, I)|T]):-
    !, 
    qspace(Par, _, Space, _), 
    memberchk(point(I), Space), 
    check_quantity_spaces(T).

check_quantity_spaces([H|_]):-
        writef('**** Error: can not apply %w\n', [H]), 
        fail.


value_sub_set([], _).
value_sub_set([value(Nm, _Q, V, D)|T], L):-
    V == unknown, 
    common_select(L, value(Nm, _QX, VX, DX), NL), 
    var(VX), nonvar(DX), 
    D = DX, 
    value_sub_set(T, NL).
value_sub_set([value(Nm, _Q, V, D)|T], L):-
    common_select(L, value(Nm, _QX, VX, DX), NL), 
    nonvar(VX), nonvar(DX), 
    V = VX, D = DX, 
    value_sub_set(T, NL).


/* FL 3-3-04, Old relations_sub_set before adding new correspondence primitives of garp 2.0***

relations_sub_set([], _, _).


% any form of correspondence between two parameters
% will produce multiple solutions if more than one 
% correspondence relation is specified

relations_sub_set([correspondence(P1, V1, P2, V2)|T], L, VList):-
    !, 
    (   common_select(L, q_correspondence(P1, P2), NL)
    ;
        common_select(L, q_correspondence(P2, P1), NL)
    ;
        common_select(L, dir_q_correspondence(P1, P2), NL)
    ;
        common_select(L, dir_q_correspondence(P2, P1), NL)
    ;
        common_select(L, v_correspondence(P1, V1, P2, V2), NL)
    ;
        common_select(L, v_correspondence(P2, V2, P1, V1), NL)
    ;
        common_select(L, dir_v_correspondence(P1, V1, P2, V2), NL)
    ;
        common_select(L, dir_v_correspondence(P2, V2, P1, V1), NL)
    ), 
    relations_sub_set(T, NL, VList ).

relations_sub_set([H|T], L, VList ):-
    (   common_select(L, H, NL)
    ;
        alternative_relation(H, IH), 
        common_select(L, IH, NL)
        ;
                try_value_consistency( H, L, VList, NL )
    ), 
    relations_sub_set(T, NL, VList).
********* GARP 2.0 version of  relations_sub_set below: ***/

/* new correspondence primitives reference:

- dv_correspondence(_, _, _, _)
- dq_correspondence(_, _)
- dir_dv_correspondence(_, _, _, _) 
- dir_dq_correspondence(_, _)
- full_correspondence(_, _)
- dir_full_correspondence(_, _)
- mirror_q_correspondence(_, _)
- mirror_dq_correspondence(_, _)

NB!
Since the rule interpreter that uses this procedure aspects only 
correspondence between parameters, all derivative correspondences
are deactivated in this procedure.
*/

relations_sub_set([], _, _).

% any form of correspondence between two parameters
% will produce multiple solutions if more than one 
% correspondence relation is specified

relations_sub_set([correspondence(P1, V1, P2, V2)|T], L, VList):-
    !, 
    (
    member(Corr, [  q_correspondence(P1, P2),
                    q_correspondence(P2, P1),
                    dir_q_correspondence(P1, P2),
                    dir_q_correspondence(P2, P1),
                    v_correspondence(P1, V1, P2, V2),
                    v_correspondence(P2, V2, P1, V1), 
                    dir_v_correspondence(P1, V1, P2, V2),
                    dir_v_correspondence(P2, V2, P1, V1),
%                   dq_correspondence(P1, P2),
%                   dq_correspondence(P2, P1),
%                   dir_dq_correspondence(P1, P2),
%                   dir_dq_correspondence(P2, P1),
%                   dv_correspondence(P1, V1, P2, V2),
%                   dv_correspondence(P2, V2, P1, V1), 
%                   dir_dv_correspondence(P1, V1, P2, V2),
%                   dir_dv_correspondence(P2, V2, P1, V1),
                    full_correspondence(P1, P2),
                    full_correspondence(P2, P1),
%                   mirror_q_correspondence(P1, P2),
%                   mirror_q_correspondence(P2, P1),
%                   dir_mirror_q_correspondence(P1, P2),
%                   dir_mirror_q_correspondence(P2, P1),
%                   mirror_dq_correspondence(P1, P2),
%                   mirror_dq_correspondence(P2, P1),
%                   dir_mirror_dq_correspondence(P1, P2),
%                   dir_mirror_dq_correspondence(P2, P1),
                    dir_full_correspondence(P1, P2),
                    dir_full_correspondence(P2, P1)
                 ]),                    
    common_select(L, Corr, NL)
    )
    ;
    ( 
    member(Corr, [  mirror_q_correspondence(P1, P2),
                    mirror_q_correspondence(P2, P1),
                    dir_mirror_q_correspondence(P1, P2),
                    dir_mirror_q_correspondence(P2, P1)
%                   mirror_dq_correspondence(P1, P2),
%                   mirror_dq_correspondence(P2, P1),
%                   dir_mirror_dq_correspondence(P1, P2),
%                   dir_mirror_dq_correspondence(P2, P1)
                 ]),                    
    common_select(L, Corr, NL),
    mirrorable_qspaces(P1, P2, List),
    memberchk(V1/V2, List)
    ), 
    relations_sub_set(T, NL, VList ).

relations_sub_set([H|T], L, VList ):-
    (   common_select(L, H, NL)
    ;
        alternative_relation(H, IH), 
        common_select(L, IH, NL)
        ;
                try_value_consistency( H, L, VList, NL )
    ), 
    relations_sub_set(T, NL, VList).



% Part of REPAIR: 23/08/2001 by BB

/*
All the rules for the transition step require a 100% match with what is
present in the state. So if a 5 valued quantity has it lowest value and a
constraint requires that the quantity is lower then its highest value, this
constraint will usually not succeed. Although consistent, this fact is
not directly present in the state. Here a procedure is added that does
allow rules to evaluate such constraints. If the constraint cannot be
match directly, but is consistent with the current state: proceed as if
rule is valid. However, the procedure is limited in that it only works
for values constraints (not for constraints on derivatives) and that it
does not do any transitivity reasoning. For the time being that seems
to be sufficient... */

% ALL NEW

try_value_consistency( H, _L, VList, _NL ):-
          % must be a in equality relation between a quantity (LHS) and a
          % value (RHS) no variables, only instances of quantities and values
          % the constraint value must exist in the Qspace (and should
          % be a point value)
          H =.. [ Rel, LHS, RHS ],
          % should we include equal and inequalities for derivatives...
          member( Rel, [greater, smaller, greater_or_equal, smaller_or_equal] ),
          nonvar( LHS ), nonvar( RHS ),
          memberchk( value( LHS, _SE, V, _D), VList ),
          qspace( LHS, _, QSpace, _ ),
          memberchk( point( RHS ), QSpace ),
          nonvar( V ),

	 (
	    @tracer->>tell(ordering,termination,transition) 
	  ->
            write('Trying consistency check: '),
            write_ln( H ),
            @tracer->>told,
            !
         ;
            true
	 ), 

          % separate QSpace wrt current value
          separate( V, QSpace, Lower, Higher ),
          % make constraint
          Constraint =.. [ Rel, RHS ],
          % test consistency
          value_consistency( Constraint, V, Lower, Higher ).


% Separate the values from a Qspace into Lower or Higher
% then the value the quantity currently has.
% Stop when Qspace becomes empty
separate( _, [], [], [] ):- !.
% When value matches, put rest QSpace in Higher list
% (match a point value)
separate( V, [point( V )|Higher], [], Higher ):- !.
% (now it should be an interval, but we don't check...)
separate( V, [V|Higher], [], Higher ):- !.
% When no match, put H1 (V in Qspace) in Lower list
separate( V, [H1|T1], [H1|T2], List ):-
        separate( V, T1, T2, List ).

value_consistency( equal( V ), V, _, _ ):- !.
value_consistency( greater( V ), _, Lower, _ ):-
        ( memberchk( point( V ), Lower ) ; memberchk( V, Lower ) ).
value_consistency( smaller( V ), _, _, Greater ):-
        ( memberchk( point(V), Greater ) ; memberchk( V, Greater ) ).
value_consistency( greater_or_equal( V ), V, _, _ ):- !.
value_consistency( greater_or_equal( V ), _, Lower, _ ):-
        ( memberchk( point( V ), Lower ) ; memberchk( V, Lower ) ).
value_consistency( smaller_or_equal( V ), V, _, _ ):- !.
value_consistency( smaller_or_equal( V ), _, _, Higher ):-
        ( memberchk( point(V), Higher ) ; memberchk( V, Higher ) ).

% End of REPAIR: 23/08/2001 by BB


% corresponding values for a parameter:
% values must be at same place in qspaces

get_corresponding_values([H1|_], [H2|_], V1, V2):-
    once((H1 = point(V1) ; H1 = V1)), 
    once((H2 = point(V2) ; H2 = V2)).

get_corresponding_values([_|T1], [_|T2], V1, V2):-
    get_corresponding_values(T1, T2, V1, V2).

% alternative ways to specify relations:

alternative_relation(v_correspondence(P1, V1, P2, V2), 
              v_correspondence(P2, V2, P1, V1)).
alternative_relation(q_correspondence(L, R), q_correspondence(R, L)).
alternative_relation(smaller(L, R), greater(R, L)).
alternative_relation(greater(L, R), smaller(R, L)).
alternative_relation(equal(L, R), equal(R, L)).
alternative_relation(smaller_or_equal(L, R), greater_or_equal(R, L)).
alternative_relation(greater_or_equal(L, R), smaller_or_equal(R, L)).
% FL 3-3-04: added derivative relations...
alternative_relation(d_smaller(L, R), d_greater(R, L)).
alternative_relation(d_greater(L, R), d_smaller(R, L)).
alternative_relation(d_equal(L, R), d_equal(R, L)).
alternative_relation(d_smaller_or_equal(L, R), d_greater_or_equal(R, L)).
alternative_relation(d_greater_or_equal(L, R), d_smaller_or_equal(R, L)).
% FL 3-3-04: NEW correspondence primitives added: 
alternative_relation(dv_correspondence(P1, V1, P2, V2), 
              dv_correspondence(P2, V2, P1, V1)).
alternative_relation(dq_correspondence(L, R), dq_correspondence(R, L)).
alternative_relation(full_correspondence(L, R), full_correspondence(R, L)).
alternative_relation(mirror_q_correspondence(L, R), mirror_q_correspondence(R, L)).
alternative_relation(mirror_dq_correspondence(L, R), mirror_dq_correspondence(R, L)).

/* correspondence primitives reference:
- dv_correspondence(_, _, _, _)
- dq_correspondence(_, _)
- dir_dv_correspondence(_, _, _, _) 
- dir_dq_correspondence(_, _)
- full_correspondence(_, _)
- dir_full_correspondence(_, _)
- mirror_q_correspondence(_, _)
- dir_mirror_q_correspondence(_, _)
- mirror_dq_correspondence(_, _)
- dir_mirror_dq_correspondence(_, _)
*/


% merge a list of values into another list of values
% note that if   value(Instance, Q1, V1, D1) does not match with
% value(Instance, Q2, V2, D2), the merge is not possible

merge_values([], L2, L2).

merge_values([value(Instance, Q1, V1, D1) | T1], L2, 
          [value(Instance, Q1, V1, D1) | T3]):-
    common_select(L2, value(Instance, Q2, V2, D2), T2), 
    !, 
    (
      value(Instance, Q1, V1, D1) = value(Instance, Q2, V2, D2)
    -> 
      merge_values(T1, T2, T3)
    ;  
      etrace(methods_value_clash, 
	     [value(Instance, Q1, V1, D1), value(Instance, Q2, V2, D2) ],
	     general),
      fail
    ).

merge_values([V1|T1], L2, [V1|T3]):-
    merge_values(T1, L2, T3).

% remove system elements from smd: remove same class and/or sub class

remove_system_elements([], NL, NL).

remove_system_elements([H|T], L, NL):-
    isa_instance_select(H, L, L1),  % probably succeeds once
    !,                  % but retry for safety
    remove_system_elements([H|T], L1, NL).

remove_system_elements([H|T], L, NL):-
    isa_attribute_select(H, L, L1),     % probably succeeds once
    !,                  % but retry for safety
    remove_system_elements([H|T], L1, NL).

remove_system_elements([H|T], L, NL):-
    common_select(L, H, L1),    % probably succeeds once
    !,                  % but retry for safety
    remove_system_elements([H|T], L1, NL).

remove_system_elements([_|T], L, NL):-
    remove_system_elements(T, L, NL).


% ------------------------Determine Exogenous Derivatives----------------------------
% 
% exogenous variables get derivatives before influence resolution:
% - check if parameter is in givens of exogenous MF,
% - then use instance information from Cio not from MF to determine most
% specific class.
% 
% 5 types:
% - continuous: derivative always plus on the way up, always min on the way down,
% except in highest/lowest points,  terminations arrange stopping and turning around.
% if way up/down is not defined coming from the input system, then generate all
% possibilities: up/down, zero in endpoints.
% - free: if not defined, generate all possibilitiesin lowest point
% - increasing: derivative always plus except in highest point
% - steady: derivative always zero
% - decreasing: derivative always min except in lowest point

determine_exogenous_derivatives(Cin, SS, Cout):-
    etrace(methods_det_exo_start, _, add_relation),
    cio_oe(Cin, SE),
    determine_exogenous_derivatives(SS, SE, Cin, Cout),
    etrace(methods_det_exo_done, _, add_relation).


% for every model fragment check if exogenous, 
% and act accordingly upon parameters
determine_exogenous_derivatives([], _, Cin, Cin).

% exogenous    
determine_exogenous_derivatives([SS|T], SE, Cin, Cout):-
    is_exogenous_mf(SS, SE, Type, Par),
    !,
    value_low_middle_high(Par, Cin, VLMH),  
    determine_exogenous_derivative(Par, Type, Cin, Cnew, VLMH),
    determine_exogenous_derivatives(T, SE, Cnew, Cout).

% not exogenous
determine_exogenous_derivatives([_|T], SE, Cin, Cout):-
   determine_exogenous_derivatives(T, SE, Cin, Cout).

% value_low_middle_high(+Par, +Cin, -VLMH)
% determine if par is in lowest, highest or a middle value.
% return values: low(interval/point), middle, high(interval/point)
value_low_middle_high(Par, Cin, VLMH):-  
    cio_d(Cin, Derivable),
    qspace(Par, _, Space, _), 
    get_derivable_value(value(Par), Interval, Space, Cin, _, 
            Derivable), 
    low_middle_high(Space, Interval, VLMH).


low_middle_high([Low|_], Low, low(interval)):-!.

low_middle_high([point(Low)|_], Low, low(point)):-!.

low_middle_high(Space, High, high(interval)):-
    last(Space, High),
    !.

low_middle_high(Space, High, high(point)):-
    last(Space, point(High)),
    !.

low_middle_high(_, _, middle).

    
% determine_exogenous_derivative(+Par, +Type, +Cin, -Cnew, VLMH)
% set derivatives:
% continuous & free can be backtrackable, others not.

% exogenous_sinus: sign known  
determine_exogenous_derivative(Par, exogenous_sinus, Cin, Cin, VLMH):-
    cio_q(Cin, Q),
    cio_d(Cin, D),
    memberchk(derivative(Par)/I, Q),
    list_map([I], Map),
    get_quantity_sign(Map, D, Sign),
    !,
    (
    (Sign = zero, VLMH = middle) 
    -> 
    (etrace(methods_sinus_zero_in_middle_error, Par, general), fail) 
    ; 
    true
    ). 
    % in middle values sinus may not be zero. Correspondences can wrongfully set derivative 
        
% exogenous_sinus: sign unknown -> new from inputsystem: 
% return all possibilities: in middle value: plus and min
determine_exogenous_derivative(Par, exogenous_sinus, Cin, Cout, middle):-
    List = [Par, exogenous_sinus, Rel, middle, input, Nr, 2],
    etrace(methods_set_exo_der_branchepoint, List, general),
    !,
    member(Rel, [d_smaller(Par, zero), d_greater(Par, zero)]),
    nth1(Nr, [d_smaller(Par, zero), d_greater(Par, zero)], Rel),
    etrace(methods_set_exo_der_intobranche, List, resolve), % tempflo: cat? 
    intern_representation(Rel, Relation, Cin, Cnew),
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true).

% exogenous_sinus: sign unknown -> new from inputsystem: 
% return all possibilities: in extreme values: all
determine_exogenous_derivative(Par, exogenous_sinus, Cin, Cout, _):-
    List = [Par, exogenous_sinus, Rel, extreme, input, Nr, 3],
    etrace(methods_set_exo_der_branchepoint, List, general),
    !, % tempflo new cut march 07
    member(Rel, [d_smaller(Par, zero), d_equal(Par, zero), d_greater(Par, zero)]),
    nth1(Nr, [d_smaller(Par, zero), d_equal(Par, zero), d_greater(Par, zero)], Rel),
    etrace(methods_set_exo_der_intobranche, List, resolve), % tempflo: cat? 
    intern_representation(Rel, Relation, Cin, Cnew),
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true).

% exogenous_free: sign already known
determine_exogenous_derivative(Par, exogenous_free, Cin, Cin, _):-
    cio_q(Cin, Q),
    cio_d(Cin, D),
    memberchk(derivative(Par)/I, Q),
    list_map([I], Map),
    get_quantity_sign(Map, D, _),
    !.

% exogenous_free: sign unknown -> new from inputsystem: 
% return all possibilities
determine_exogenous_derivative(Par, exogenous_free, Cin, Cout, _):-
    List = [Par, exogenous_free, Rel, nil, input, Nr, 3],
    etrace(methods_set_exo_der_branchepoint, List, general),
    !, % tempflo new cut march 07
    member(Rel, [d_smaller(Par, zero), d_equal(Par, zero), d_greater(Par, zero)]),
    nth1(Nr, [d_smaller(Par, zero), d_equal(Par, zero), d_greater(Par, zero)], Rel),
    intern_representation(Rel, Relation, Cin, Cnew),
    etrace(methods_set_exo_der_intobranche, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true).

% 22-2-2007 FL: reinserted clause, 
% thus changed from old semantics 20-8-2004 FL:  exogenous increasing should always be dPlus,
% exogenous_increasing: arrived at highest point
determine_exogenous_derivative(Par, exogenous_increasing, Cin, Cout, high(point)):-
    List = [Par, exogenous_increasing, d_equal(Par, zero), highestpoint, nil, nil, nil],
    intern_representation(d_equal(Par, zero), Relation, Cin, Cnew),
    etrace(methods_set_exo_der, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true),
    !.


% exogenous_increasing: set to plus
determine_exogenous_derivative(Par, exogenous_increasing, Cin, Cout, _):-
    List = [Par, exogenous_increasing, d_greater(Par, zero), nil, nil, nil, nil],
    intern_representation(d_greater(Par, zero), Relation, Cin, Cnew),
    etrace(methods_set_exo_der, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true),
    !.

% exogenous_steady: already known zero
determine_exogenous_derivative(Par, exogenous_steady, Cin, Cin, _):-
    cio_q(Cin, Q),
    cio_d(Cin, D),
    memberchk(derivative(Par)/I, Q),
    list_map([I], Map),
    get_quantity_sign(Map, D, Sign),
    Sign = zero,
    !.

% exogenous_steady: set to zero
determine_exogenous_derivative(Par, exogenous_steady, Cin, Cout, _):-
    List = [Par, exogenous_increasing, d_equal(Par, zero), nil, nil, nil, nil],
    intern_representation(d_equal(Par, zero), Relation, Cin, Cnew),
    etrace(methods_set_exo_der, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true),
    !.

% 22-2-2007 FL: reinserted clause,  
% thus changed from old semantics of 20-8-2004 : decr. should always be dMin
% exogenous_decreasing: arrived at lowest point
determine_exogenous_derivative(Par, exogenous_decreasing, Cin, Cout, low(point)):-
    intern_representation(d_equal(Par, zero), Relation, Cin, Cnew),
    List = [Par, exogenous_decreasing, d_equal(Par, zero), lowestpoint, nil, nil, nil],
    etrace(methods_set_exo_der, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true),
    !.


% exogenous_decreasing: set to min
determine_exogenous_derivative(Par, exogenous_decreasing, Cin, Cout, _):-
    intern_representation(d_smaller(Par, zero), Relation, Cin, Cnew),
    List = [Par, exogenous_decreasing, d_smaller(Par, zero), nil, nil, nil, nil],
    etrace(methods_set_exo_der, List, resolve), % tempflo: cat? 
    exogenous_try_append_relation(List, Relation, Cnew, Cout, _, true),
    !.

% extra layer for tracer
exogenous_try_append_relation(_, Relation, Cnew, Cout, _, true):-
    append_relation(Relation, Cnew, Cout, _, true),
    !.
exogenous_try_append_relation(List, _, _, _, _, _):-
    etrace(methods_set_exo_der_fail, List, general), % tempflo: cat? 
    fail.


% a parameter appears in givens of a MF which requires
% exogenous assumption. Return type of exogenous variable
is_exogenous_mf(SS, SElist, Type, Par):-
    system_structure_slots(SS, _, _, C, G),
    memberchk(system_elements(SEfromSS), C),
    member(instance(Assumption, Class), SEfromSS),
    isa_instance(Class, exogenous),
    !, % SS is exogenous, now determine specific type from seList, MF may be general
    specific_exogenous_type(Assumption, SElist, Type),
    memberchk(parameters([P]), G), % exogenous MF should introduce only one parameter
    P =..[_, _, Par, _, _].

% exogenous_sinus
specific_exogenous_type(Assumption, SElist, exogenous_sinus):-
    member(instance(Assumption, Class), SElist),
    isa_instance(Class, exogenous_sinus),
    !.

% exogenous_free
specific_exogenous_type(Assumption, SElist, exogenous_free):-
    member(instance(Assumption, Class), SElist),
    isa_instance(Class, exogenous_free),
    !.

% exogenous_increasing
specific_exogenous_type(Assumption, SElist, exogenous_increasing):-
    member(instance(Assumption, Class), SElist),
    isa_instance(Class, exogenous_increasing),
    !.

% exogenous_steady
specific_exogenous_type(Assumption, SElist, exogenous_steady):-
    member(instance(Assumption, Class), SElist),
    isa_instance(Class, exogenous_steady),
    !.

% exogenous_decreasing
specific_exogenous_type(Assumption, SElist, exogenous_decreasing):-
    member(instance(Assumption, Class), SElist),
    isa_instance(Class, exogenous_decreasing),
    !.

