/*  File:    solve
    Purpose: Inequality reasoning
    Author:  Martin Reinders & Bert Bredeweg & Floris Linnebank
    Date:    August 1989
    Part-of:  GARP (version 2.0)
    Modified: 30 July 2004

    Copyright (c) 2004, University of Amsterdam. All rights reserved.

*/


/*
 *  is_derivable/2
 *
 *  1 + Relation, intern representation (<, =< reversed to >, >=)
 *  2 + Relations, known to be derivable in the current context (idem)
 *
 *  is 1 a member of 2? (or directly implied by 2?, e.g. a > b -> a >= b
 *
 */

is_derivable(Relation, _):- evident(Relation), !.

is_derivable(relation(Left, Relation, Right), Derivable):-
	memberchk(relation(Left, Relation, Right), Derivable), !.
% inverse for equality relation
is_derivable(relation(Left, =, Right), Derivable):-
	memberchk(relation(Right, =, Left), Derivable), !.
is_derivable(relation(Left, Relation, Right), Derivable):-
	implies_relation(Implies, Relation),
	% FL july 2004: implies can be = and should be checked both ways: 
	% memberchk(relation(Left, Implies, Right), Derivable), !.  % old
	is_derivable(relation(Left, Implies, Right), Derivable), !. % new

/*
 * is a list of relations (intern representation) completely derivable?
 */

check_derivable([], _).
check_derivable([H|T], D):-
	is_derivable(H, D),
	!,
	check_derivable(T, D).

/*
 *  implies_relation(+op1, -op2)
 *  a op1 b  ->  a op2 b
 *
 */

implies_relation(>, >=).
implies_relation(=, >=).

/*
 * append_relation(+Relation, +Cin, -Cout, -Appended, +AnalyseSimple)
 * append_relation_all(+RelationList, .. idem ..).
 * append a relation/relationlist to the cio structure
 * Relation in intern representaion
 * Appended: relations appended (i.e. not already derivable)
 * AnalyseSimple: true -> replace a with b if a=b is derived
 *		  false -> don't replace
 *
 */

% FL december 2004:
% In the influence resolution procedure in garp 2.0 the list of adders
% (influences to add) is passed on  when appending a relation. 
% This way, AnalyseSimpleEquality and Zero Analysis can
% change pointers/bits, without invalidating the adders.

% other calls do not yet use these arguments so this extra clause
% is added to catch old style calls.
append_relation(R, Cin, Cout, Ap, As):-     
    append_relation(R, Cin, Cout, Ap, As, [],[]).


% FL december 2004: Solve does simplification of results:
% x + y = x + z  ->  y = z,
% but not simplification of the input... This is done here, before
% the actual appendrelation2 call. 
% Last two arguments are Adders passed on for bit change updates. 
% (see: analyse zero equality)
append_relation(R, Cin, Cout, Ap, As, Addersin, Addersout):-
     simplify_relations([R], [R1]),     
    !,
    append_relation2(R1, Cin, Cout, Ap, As, Addersin, Addersout).


% fail if relation is invalid (i.e. a > a, or a+b > a+b)
% FL december 2004: new name & 2 extra arguments, see above.

append_relation2(Intern, _, _, _, _, _, _):- 
	illegal_relation(Intern),
	% etrace(solve_found_invalid, _, inequality), gives too many extra statements in e.g. Inf Res.
	!,
	fail.

% relation is already derivable
% FL december 2004: new name & 2 extra arguments, see above.

append_relation2(Intern, Cio, Cio, [], _, Ad, Ad):- 
	cio_d(Cio, Derivable),
	is_derivable(Intern, Derivable),
	% etrace(solve_derivable_rel, _, add_relation), gives too many extra statements in e.g. Inf Res.
	!.

% append relation (fail if inconsistent)
% FL december 2004: new name (+2) & 2 extra arguments, see above.

append_relation2(AppendOne, Cin, Cout, [AppendOne], DoAnalyseSimple, Adders, NAdders):-
	cio_r_nr_d_nd_q_nq_c_nc(Cin, Cnew,
		Relations, SortedFinalAllRelations,
		Derivable, SortedFinalAllNewDerivable,
		Q, NQ,
		C, NC),
	append([AppendOne], Derivable, KnownDerivable),
	% all relations we don't want to derive again
	solve([[]/AppendOne], Relations, Relations, KnownDerivable, NewDerived, Q),
	% fails is inconsistent
	append(KnownDerivable, NewDerived, AllNewDerivable),
	% all relations we have derived now
	append([AppendOne], NewDerived, AllNew),
	% all new relations
	append([AppendOne], Relations, AllRelations),
	% all relations used as a base
	remove_weaker(AllNew, AllNew, StrongAllNew),
	% remove a>=b if a>b is found
	remove_weaker(StrongAllNew, AllRelations, StrongAllRelations),
	% also in base
	remove_weaker(StrongAllNew, AllNewDerivable, StrongAllNewDerivable),
	% also in all derived
	flag(analyse_simple_found_contradiction, _, false),
	analyse_simple_equality(DoAnalyseSimple, StrongAllNew, Q, Q1,
		StrongAllNew, SimpleAllNew,
		StrongAllRelations, SimpleAllRelations,
		StrongAllNewDerivable, SimpleAllNewDerivable,
		C, C1, Adders, Adders1), % Note that adderlists are passed on
	% replace a with b if a=b
	analyse_zero_equality(DoAnalyseSimple, SimpleAllNew, FinalAllNew, 
	    SimpleAllNewDerivable, FinalAllNewDerivable, 
	    SimpleAllRelations, FinalAllRelations,
	    Q1, NQ, C1, NC, Adders1, NAdders), % Note that adderlists are passed on     
	% FL december 2004 new in garp 2.0: zero analysis = analyse simple equality using: 
	% x = 0, y = 0 -> x=y pattern
	sort(FinalAllRelations, SortedFinalAllRelations),
	% remove duplicates
	sort(FinalAllNew, SortedFinalAllNew),
	sort(FinalAllNewDerivable, SortedFinalAllNewDerivable),
	!,
	inspect_correspondence(SortedFinalAllNew, Cnew, Cout, DoAnalyseSimple).
	% inspect if and iff statements (correspondence)


% FL december 2004: 2 extra arguments, see above.
append_relation_all(Append, Cin, Cout, AppendNotDerivable, DoAnalyseSimple):-
    append_relation_all(Append, Cin, Cout, AppendNotDerivable, DoAnalyseSimple, [], []).

% FL december 2004: Solve does simplification of results:
% x + y = x + z  ->  y = z,
% but not simplification of the input... This is done here, before
% the actual appendrelation2 call. 
% Last two arguments are Adders passed on for bit change updates. 
% (see: analyse zero equality)
append_relation_all(Append1, Cin, Cout, AppendNotDerivable, DoAnalyseSimple, Adin, Adout):-
	simplify_relations(Append1, Append),
	!,
	append_relation_all2(Append, Cin, Cout, AppendNotDerivable, DoAnalyseSimple, Adin, Adout).


% fail if invalid input

% FL december 2004: new name & 2 extra arguments, see above.
append_relation_all2(InternL, _, _, _, _, _, _):- 
	member(Intern, InternL),
	illegal_relation(Intern),
	!,
	fail.

% append relations (fail if inconsistent)
% Adderlist is passed on for bit change updates: see analyse zero equality
% results in 2 extra arguments

append_relation_all2(Append, Cin, Cout, AppendNotDerivable, DoAnalyseSimple, A, NA):- 
	cio_r_nr_d_nd_q_nq_c_nc(Cin, Cnew,
		Relations, SortedFinalAllRelations,
		Derivable, SortedFinalAllNewDerivable,
		Q, NQ,
		C, NC),
	remove_derivable(Append, Derivable, AppendNotDerivable),
	!,
(	AppendNotDerivable == [], Cin = Cout, A = NA
	;
	append(AppendNotDerivable, Derivable, KnownDerivable),
	% all relations we don't want to derive again
	append(AppendNotDerivable, Relations, KnownRelations),
	add_empty_parents(AppendNotDerivable, AppendNotDerivable2),
	solve(AppendNotDerivable2, KnownRelations, KnownRelations,
			KnownDerivable, NewDerived, Q),
	% fails is inconsistent
	append(KnownDerivable, NewDerived, AllNewDerivable),
	% all relations we have derived now
	append(AppendNotDerivable, NewDerived, AllNew),
	% all new relations
	append(AppendNotDerivable, Relations, AllRelations),
	% all relations used as a base
	remove_weaker(AllNew, AllNew, StrongAllNew),
	% remove a>=b if a>b is found
	remove_weaker(StrongAllNew, AllRelations, StrongAllRelations),
	% also in base
	remove_weaker(StrongAllNew, AllNewDerivable, StrongAllNewDerivable),
	% also in all derived
	analyse_simple_equality(DoAnalyseSimple, StrongAllNew, Q, NQ,
		StrongAllNew, FinalAllNew,
		StrongAllRelations, FinalAllRelations,
		StrongAllNewDerivable, FinalAllNewDerivable,
		C, NC, A, NA), % Note that adderlists are passed on 
	% replace a with b if a=b
	% analyse zero equality is not done here, because it is done in the
	% single version of append relation, 
    % because it is exhaustive it is most efficient to do it once in a while.
	sort(FinalAllRelations, SortedFinalAllRelations),
	% remove duplicates
	sort(FinalAllNew, SortedFinalAllNew),
	sort(FinalAllNewDerivable, SortedFinalAllNewDerivable),
	!,
	inspect_correspondence(SortedFinalAllNew, Cnew, Cout, DoAnalyseSimple)
	% inspect if and iff statements
 ),
	!.


/*
 * remove_derivable(+ToInspect, +KnownRelations, -NotDerivable)
 *
 * remove all derivable relations (given Arg2) from Arg1
 *
 */

remove_derivable([], _, []).

remove_derivable([H|T], D, NT):-
	is_derivable(H, D),
	!,
	remove_derivable(T, D, NT).

remove_derivable([H|T], D, [H|NT]):-
	remove_derivable(T, D, NT).

/*
 * remove_weaker(+Newrelations, +KnownRelations, -KnownStronger)
 *
 * Arg1 is a list of new (derived) relations
 * Arg2 is a list of known (derived) relations
 * remove from Arg2 each relation a >= b if Arg1 contains a > b/ a = b
 *
 */

remove_weaker([ relation(L, >, R) | Tail], Rs, NRs):-
	memberchk(relation(L, >=, R), Rs),
	% memberchk is logically superfluous,
	% but, while written in C, the test is faster
	common_select(Rs, relation(L, >=, R), TRs), !,
	remove_weaker([ relation(L, >, R) | Tail], TRs, NRs).
remove_weaker([ relation(L, =, R) | Tail], Rs, NRs):-
	memberchk(relation(L, >=, R), Rs),
	common_select(Rs, relation(L, >=, R), TRs), !,
	remove_weaker([ relation(L, =, R) | Tail], TRs, NRs).
remove_weaker([ relation(R, =, L) | Tail], Rs, NRs):-
	memberchk(relation(L, >=, R), Rs),
	common_select(Rs, relation(L, >=, R), TRs), !,
	remove_weaker([ relation(L, =, R) | Tail], TRs, NRs).
remove_weaker([_|Tail], Rs, NRs):-
	remove_weaker(Tail, Rs, NRs).
remove_weaker([], Rs, Rs).

/*
 * illegal_relations(+Relation)
 *
 * succeeds if Relation (intern) is a + term > a + term
 *
 */

illegal_relation(relation(Left, Relation, Right)):-
	map_without_intersection(Left, Right, NLeft, NRight),
	% fails if no intersection
	invalid_relation(relation(NLeft, Relation, NRight)),
	!.

illegal_relation(relation(Left, Relation, Right)):-
	invalid_relation(relation(Left, Relation, Right)),
	!.

invalid_relation(relation(M, >, M)).

add_empty_parents([], []).
add_empty_parents([H|T], [[]/H|NT]):- add_empty_parents(T, NT).


/* --------------------------------------------------------------------
 * solve(+NewRelations, +ToInspect, +AllKnown, +AllDerivable, -New)
 *
 * for each relation in NewRelations:
 * add first of NewRelations with all of ToInspect (= AllKnown on first call)
 * if a valid, not evident and not already known relation is found:
 * add it to NewRelations
 * Return only relations between two single quantities or constants
 *
 */

solve(NR, TO, K, D, New, Q):-
	@tracer->>trace_on(inequality), % FL Feb 07 tell and told should be in print routines!
					% because solve_trace fails in case of contradiction!
					% leads to traces being 'told' much later...
	!,
	solve_trace(NR, TO, K, D, New, Q). 


solve(NR, TO, K, D, New, _):-
	solve(NR, TO, K, D, New).

% done, return result

solve([], _, _, _, []):- !.


% add relation, succeeds if result is consistent,
% has no term twice and can be simplified
% if inconsistent: fail solve
% if no simplification, or already derivable, or evident: next clause

solve([Parents/Relation1|TailNew], KnownRelations, Known, Derivable, New):-
	member_and_tail(Relation2, KnownRelations, TailKnown),
	add_relation(Relation1, Relation2, Relation3),  % transivity
	\+ memberchk(Relation2, Parents), 		 % not circular 
	\+ is_derivable(Relation3, Derivable), 	 % not derivable
	\+ evident(Relation3), 				 % not evident
	!,
	\+ invalid_relation(Relation3), 		 % fail if contradiction
	return_simple_only(Relation3, NNew, New),
	!,
	solve([Parents/Relation1, [Relation1, Relation2|Parents]/ Relation3
		| TailNew], TailKnown, Known, [Relation3|Derivable], NNew).

% tried all for first relation, try tail

solve([_|TailNew], _, Known, Derivable, New):-
	!,
	solve(TailNew, Known, Known, Derivable, New).


/* ------------------- same when tracing inequality reasoning ---------- */

% done, return result

solve_trace([], _, _, _, [], _):- !.


% add relation, succeeds if result is consistent,
% has no term twice and can be simplified
% if inconsistent: fail solve
% if no simplification, or already derivable, or evident: next clause

solve_trace([Parents/Relation1|TailNew], KnownRelations, Known, Derivable, New, Q):-
	member_and_tail(Relation2, KnownRelations, TailKnown),
	add_relation(Relation1, Relation2, Relation3),  % transivity
	\+ memberchk(Relation2, Parents), 		 % not circular
	\+ is_derivable(Relation3, Derivable), 	 % not derivable
	\+ evident(Relation3), 				 % not evident
	!,
	(invalid_relation(Relation3) ->			 % fail if contradiction
		etrace(solve_derived_invalid, _, inequality),
	        @tracer->>tell(inequality),	 
	        show_antecedents(Parents, Relation1, Relation2, Relation3, 'contradiction.', Q),
	        @tracer->>told,
		fail
		;
		true
	),
	return_simple_only(Relation3, NNew, New),
	!,
	(simple_relation(Relation3) ->
	  etrace(solve_found_valid, _, inequality),
	  @tracer->>tell(inequality),	 
	  show_antecedents(Parents, Relation1, Relation2, Relation3, 'derived.', Q),
	  @tracer->>told
	  ;
	  true
	),
	append(Parents, [Relation1, Relation2], NewParents),
	solve_trace([Parents/Relation1, NewParents/ Relation3 | TailNew],
			 TailKnown, Known, [Relation3|Derivable], NNew, Q).

% tried all for first relation, try tail

solve_trace([_|TailNew], _, Known, Derivable, New, Q):-
	!,
	solve_trace(TailNew, Known, Known, Derivable, New, Q).



%%	FL Feb 2007: Moved show_antecedents to interface_garp3.pl



/* --------------------------------------------------------------------- */

member_and_tail(H, [H|T], T).
member_and_tail(H, [_|T], NT):- member_and_tail(H, T, NT).

% return only parameter > landmark, parameter > parameter, etc.

return_simple_only(Relation, New, [Relation|New]) :-
	simple_relation(Relation),
	!.

return_simple_only(_, New, New).

/*
 * add_relation(+R1, +R2, -R3)
 * add relation between two sums, represented as bitmaps
 *
 * succeeds only if both left sides or right sides don't have a common term
 * and the result can be simplified
 *
 * This scheme implements transitivity and reasoning about sums
 *
 * some examples:
 *
 * a > b		/\ b > c		-> a > c
 * a + b > c		/\ a = d		-> d + b > c
 *	(this result is not returned, but may lead to another result)
 * d = 0		/\ d + b > c		-> b > c
 *
 * note a lot of relations can be derived in many ways
 *
 * to prevent some superfluous derivations, if a quantity is equal to
 * another quantity, it is replaced with that quantity (if possible)
 * see append_relations(_all) / analyse_simple_equality
 *
 */

% Patch 3-3-2004 by FL: this case wasn't captured, because it cannot be 
% derived using transitivity. This technique of combining weaker relations
% to stronger ones is mentioned in simmons '89,
add_relation(relation(X, >=, Y), relation(Y, >=, X), relation(X, =, Y)):-
	!.


% transitivity:

add_relation(relation(Le1, Rel1, Ri1), relation(Le2, Rel2, Ri2),
		relation(Le3, Rel3, Ri3)):-
	add_compatible(Rel1, Rel2, Rel3),
	map_union_unique(Le1, Le2, L), 	  % union left sides (no term twice)
	map_union_unique(Ri1, Ri2, R),
	map_without_intersection(L, R, Le3, Ri3),  % simplify
	!.

% if first clause failed, and one relation is '=', then reverse one relation

add_relation(relation(Le1, Rel1, Ri1), relation(Ri2, =, Le2),
		relation(Le3, Rel3, Ri3)):-
	!, 	% do not use clause 3
	add_compatible(Rel1, =, Rel3),
	map_union_unique(Le1, Le2, L),
	map_union_unique(Ri1, Ri2, R),
	map_without_intersection(L, R, Le3, Ri3),
	!.

add_relation(relation(Ri1, =, Le1), relation(Le2, Rel2, Ri2),
		relation(Le3, Rel3, Ri3)):-
	add_compatible(=, Rel2, Rel3),
	map_union_unique(Le1, Le2, L),
	map_union_unique(Ri1, Ri2, R),
	map_without_intersection(L, R, Le3, Ri3),
	!.


simple_relation(relation(L, _, R)):-
	map_count(L, C1),
	C1 < 2,
	map_count(R, C2),
	C2 < 2.

evident(relation(X, =, X) ).
evident(relation(X, >=, X) ).

% < =< were inversed

add_compatible(=, =, =).
add_compatible(=, >, >).
add_compatible(=, >=, >=).
add_compatible(>, =, >).
add_compatible(>, >=, >).
add_compatible(>, >, >).
add_compatible(>=, =, >=).
add_compatible(>=, >=, >=).
add_compatible(>=, >, >).

inverse(  <, >	).
inverse(=<, >=).
inverse(  =, =	).
inverse(>=, =<).
inverse(  >, <	).


% ----------- resolve influences and proportional relations ------------

resolve_influence_proportional(Cin, Cout):-
	cio_ip_nip(Cin, Cnew, IP, []),      % collect i/p relations
	etrace(solve_ir_search, _, resolve),
	findall(adder(Quantity, AddList),   % find all effects on each quantity
		 adder(Quantity, _, AddList, [], IP, []),
		 Adders),
	!,
	resolve_adders(Adders, Cnew, Cout). % resolve these effects per quantity
    % backtracking allowed for ambiguous influences/prop. rel.

% adder/6: 
% find all influences and proportional relation on a quantity

% no more inf/prop for a certain quantity, return Addlist
adder(Quantity, Quantity, AddListOut, AddListIn, [], _):-
    nonvar(Quantity),
    list_to_set(AddListIn, AddListOut). %FL: filter double adders (bad model)

% backtrack: select next quantity and find inf/prop in restlist
adder(Quantity, _, AddList, _, [], RestIP) :-
	RestIP \== [],
	adder(Quantity, _, AddList, [], RestIP, []).

% if called for the first time Quantity is uninstantiated
% and will be matched with the quantity in prop/inf Head
% otherwise Head must be a next prop/inf on Quantity.
adder(ResultQuantity, Quantity, AddList, Previous, [Head | Tail ], RestIP) :-
	adder_relation(Head, Quantity, First),
	!,
	adder(ResultQuantity, Quantity, AddList, [First|Previous], Tail, RestIP).

% Head is not an inf/prop on Quantity, put it on RestList
adder(ResultQuantity, Quantity, AddList, Previous, [Head | Tail ], RestIP) :-
	adder(ResultQuantity, Quantity, AddList, Previous, Tail, [Head|RestIP]).


/*** FL december-2003: old resolve_adders from garp 1.7 ***
*     with partial closed world assumption 
*

resolve_adders(Adders, Cin, Cout):-
	cio_q_nq(Cin, Cnew, Q, NQ), 
	adders_quantities(Adders, NewAdders, Q, NQ), 
	!, 
	resolve_adders(NewAdders, Cnew, Cout, true).
	% fails if contradiction

% no progress see if we can set some uninfluenced influence to zero
% we must be conservative:
% set a quantity to zero that is  proportionalto / an influence on some
% other quantity, if possible. Then try to find new results.
% it is difficult to predict what will happen, because the quantity may
% have all kinds of relations to other quantities, for instance:
%
%  derivative(a) > derivative(b)	% not very likely, but possible
%  prop_pos(c, a)			% c is proportional to a
%  prop_pos(b, c)			% b is proportional to c
%  assume a to be zero
%  derive: derivative(b) > zero
%  resolve: c = zero
%  resolve: b = zero -> contradiction
%
% so if the assumption fails, try another one, but don't fail the 
% complete predicate

resolve_adders(Adders, Cin, Cout, false):-
	cio_q(Cin, Q), 
	member(Value/DQ, Q), 			% a value or derivative
	list_map([DQ], DM), 
	\+ memberchk(adder(DM, _), Adders), 	% not influenced itself
	member(adder(_, L), Adders), 		% it influences something
	(	memberchk(pos(DM), L);
		memberchk(neg(DM), L)
	), 
	cio_d(Cin, Der), 			% it's sign is not known
	\+ get_sign(pos(DM), Der, _), 
	list_map([], ZeroMap), 			% it can be assumed
	append_relation(relation(DM, =, ZeroMap), Cin, Cnew, _, false), 
	tracer(resolve, 'assuming %w is zero', [Value]), 
	resolve_adders(Adders, Cnew, Cout, true), 
	% now we are sure the assumption is safe !
	!.

resolve_adders(_, Cio, Cio, false).

resolve_adders(Adders, Cin, Cout, true):-	% first time, or progress
	resolve(Adders, Cin, Cnew, Unresolved, Progress), 
	resolve_adders(Unresolved, Cnew, Cout, Progress).

*** FL december-2003: end of old resolve_adders,  new garp 2.0 starts here ***/


% resolve_adders(+Adders, +Cin, -Cout)
% - change Adders from extern to intern representation
% - apply closed world on all quantities (under switch control)
% this means setting derivatives of unaffected quantities to zero.
% - resolve the resulting adders within Cio context.

resolve_adders(Adders, Cin, Cout):-
	cio_q_nq(Cin, Cnew1, Q, NQ),
	adders_quantities(Adders, NewAdders, Q, NQ), % convert adderlist to intern representation
	!,
	flag(cw_assumption, Flag, Flag), 
	(
	    algorithm_assumption_flag(Flag, fail) 
	-> 
	    Cnew2 = Cnew1 
	;
        % set unknown, uninfluenced, influencing quantities to zero 
	    etrace(solve_cw_start, _, resolve),
	    apply_closed_world_to_adders(NQ, NewAdders, Cnew1, Cnew2)
	),
	resolve_adders(NewAdders, Adders, Cnew2, Cout, true). % FL may 2004: pass on normal adderlist for use in tracer
    % fails if contradiction


% apply_closed_world_to_adders(+Quantities, +Adders, +Cin, -Cout)
% december-2003 by FL: 
% set every value or derivative to zero if unknown, uninfluenced
% and influencing something:
% NB: In this version it is done for all. But correspondences and complex 
% equations may cause quantities to be known later on, causing contradiction.
% Therefore, this feature can be switched off (see algorithm option switches),

apply_closed_world_to_adders([], _, Cio, Cio).

apply_closed_world_to_adders([Value/DQ|QTail], Adders, Cin, Cout):-
	list_map([DQ], DM),
	\+ memberchk(adder(DM, _), Adders), 	    % if not influenced itself
	member(adder(_, L), Adders), 		        % and it influences something
	(	memberchk(pos(DM), L);
		memberchk(neg(DM), L)                   % only pos neg, no up and down yet
	),
	cio_d(Cin, Derivable),
	\+ get_quantity_sign(DM, Derivable, _), 	% and it's sign is not known
	!,
	list_map([], ZeroMap), 			            % then it can be assumed zero
	% Analyse simple & zero equality should be false not to
	% create invalid pointers in Qlist or Adders
	append_relation(relation(DM, =, ZeroMap), Cin, Cnew, _, false),
	% this may fail if the model supplies information in contradiction with the assumption
	etrace(solve_cw_zero, [Value], resolve),
	!,
    apply_closed_world_to_adders(QTail, Adders, Cnew, Cout).

% quantity not zero-assumable: ignore it

apply_closed_world_to_adders([_|QTail], Adders, Cin, Cout):-
    apply_closed_world_to_adders(QTail, Adders, Cin, Cout).


/* --- resolve_addders/4: old garp 1.7.2  ----

% no progress, discard adders and derive unknown for these quantities.
resolve_adders(_, Cio, Cio, false). 

% first time, or progress
resolve_adders(Adders, Cin, Cout, true):-
	resolve(Adders, Cin, Cnew, [], Unresolved, Progress),
	resolve_adders(Unresolved, Cnew, Cout, Progress).
	
*** FL May 2004: garp 2.0 version (with tracer adders in normal representation) */

% no progress, discard adders and derive unknown for these quantities.
resolve_adders(_, _, Cio, Cio, false). 


% first time, or progress
resolve_adders(Adders, TrAdders, Cin, Cout, true):-
	resolve(Adders, TrAdders, Cin, Cnew, [], Unresolved, [], UnTrAd, Progress),
	resolve_adders(Unresolved, UnTrAd, Cnew, Cout, Progress).
	

/*** FL december-2003: GARP 1.7 resolve clauses not weighing influences... ***

% resolve: for each adder: determine signs of terms
% if ambiguous (at least one positive and one negative term): discard adder
% if positive, zero, or negative sum: add this result
% if not all terms are known: put adder on Unresolved list

resolve([], Cin, Cin, [], false).

resolve([ adder(Q, AddList) | Tail ], Cin, Cout, NewUnresolved, Progress) :-
	(@tracer->>tell(resolve) -> %gp3 0.3
		cio_q(Cin, QL),
		print_sum(Q, QL),
		write(' is influenced by / proportional to: '),
		forall(member(One, AddList), write_addone(One, QL)),
		nl,
		@tracer->>told,
		 !
		;
		true
	),
	cio_d(Cin, Derivable),
	sign_addlist(AddList, Derivable, first, NewAddList, Result),
	(@tracer->>tell(resolve) -> %gp3 0.3
		print_sum(Q, QL),
		write(' evaluated to be influenced by / proportional to: '),
		forall(member(One, NewAddList), write_addone(One, QL)),
		nl,
		write('which is '), write_ln(Result), 
		@tracer->>told,
		!
		;
		true
	),
	!,
	add_adder_result(Q, Result, Cin, Cnew, Progress, RestProgress,
		NewUnresolved, Unresolved, adder(Q, NewAddList)),
	% note this may fail
	resolve(Tail, Cnew, Cout, Unresolved, RestProgress).


write_addone(X, _):- var(X), !.
write_addone(pos(One), QL):- write('positive '), print_sum(One, QL), tab(2), !.
write_addone(neg(One), QL):- write('negative '), print_sum(One, QL), tab(2), !.
write_addone(One, _):- write(One), tab(2), !.

% incomplete information: add new adder to unresolved of tail,
% Let progress be progress of tail

add_adder_result(_, incomplete, Cin, Cin, Progress, Progress, [Adder|Unresolved],
	Unresolved, Adder):- !.

% valid new result: let progress be true if not already derivable

add_adder_result(Q, Sign, Cin, Cout, Progress, TailProgress,
		Unresolved, Unresolved, _):-
	zero_relation_sign(Relation, Sign),
	list_map([], Empty),
	right_order(relation(Q, Relation, Empty), Irel),
	append_relation(Irel, Cin, Cout, Added, false),
	(Added = [],
	  Progress=TailProgress;
	  Progress=true),
	!.

% ambiguous influence on Q: return all possibilities (upon backtracking)

add_adder_result(Q, ambiguous, Cin, Cout, Progress, TailProgress,
		Unresolved, Unresolved, _):-
	list_map([], Empty),
	member(Relation, [ <, =, > ]),
	right_order(relation(Q, Relation, Empty), Irel),
	once((
		append_relation(Irel, Cin, Cout, Added, false),
		(	Added = [],
			Progress=TailProgress;
			Progress=true
		)
	     )).

right_order(relation(L, <, R), relation(R, >, L)) :- !.
right_order(relation(L, =<, R), relation(R, >=, L)) :- !.
right_order(R, R):- !.



% determine sum of influences/proportional relations

sign_addlist([], _, Result, [], Result):- !.

% if already ambiguous, don't mind the tail

sign_addlist(_, _, ambiguous, [], ambiguous):- !.

% get sign, determine sum, do tail (let head of newtail be sign of head)

sign_addlist([H|T], Relations, PrevSign, [H|NewTail], Result):-
	memberchk(H, [ pos, zero, neg ]),
	!,
	sign_add(PrevSign, H, NewSign),
	sign_addlist(T, Relations, NewSign, NewTail, Result).

sign_addlist([H|T], Relations, PrevSign, [Sign|NewTail], Result):-
	get_sign(H, Relations, Sign),
	!,
	sign_add(PrevSign, Sign, NewSign),
	sign_addlist(T, Relations, NewSign, NewTail, Result).

% could not get sign: do tail (let Head of newtail be Head)
% if previous + tail is ambiguous, sign is not important
% otherwise return incomplete

sign_addlist([H|T], Relations, PrevSign, [H|NewTail], Result):-
	sign_addlist(T, Relations, PrevSign, NewTail, NewResult),
	sign_incomplete(NewResult, Result).


*** FL december-2003: New GARP 2.0 resolve clauses and subs below replace above ******

FL May 2004: added adders in normal representation for tracer use.
The pointers of intern representation are not unique to each quantity leading to the wrong
quantity being displayed and a useless trace: new below*

FL July 2004: If semi-ambiguous result is obtained:
(>=) then only two ambiguous alternatives are generated. 

*/

% resolve(+Adders, +TracerAdders, +Cin, -Cout, +[],
%         -Unresolved, +[], -UnresolvedTracerAdders, -Progress) 
% for each adder:
% - if not all values of terms are known (incomplete): 
%   put adder on Unresolved list,
%   progress for that addder = false, 
%   but true in general if other adders produce progress 
% - elseif single adder: add result, progress = true
% - else construct balance expression and evaluate:
% - if ambiguous (not enough information): 
%   add different values for Q upon backtracking, progress = true
% - if positive, zero, or negative: add this result, progress = true


% all done, return results:

resolve([], [], Cin, Cin, Unresolved, Unresolved, UTracer, UTracer, false). 


% adder incomplete -> put on Unresolved list

resolve([adder(Q, AddList)|Tail], [adder(TrQ, TrList)|TrTail], Cin, Cout, 
        Unresolved, UnresolvedOut, UTracer, UTracerOut, Progress) :-
    cio_q_d(Cin, QList, Derivable),
    % check quantities in Addlist for known values, 
    % transform pos, neg to up, down & noforce, 
    % return incomplete or complete to indicate if all values are known
    complete_adder(AddList, Derivable, NewAddList, Incomplete, TrList, TracerAddList),
    Incomplete = incomplete,
    !, %is it ok to cut here? because below it says: not this may fail... choicepoint seems unnecessary... Fl Feb 07 tempflo
    etrace(solve_ir_set, [TrQ, TrList, QList, Derivable], resolve),
    etrace(solve_ir_unres, [TracerAddList], resolve),
    % note this may fail
    resolve(Tail, TrTail, Cin , Cout, [adder(Q, NewAddList)|Unresolved], 
	        UnresolvedOut, [adder(TrQ, TrList)|UTracer], UTracerOut, Progress).


% adder complete, single influence -> evaluate & add results

resolve([adder(Q, [Influence])|Tail], [adder(TrQ, [TrInf])|TrTail], Cin, Cout, 
        UnresolvedIn, UnresolvedOut, UTracer, UTracerOut, Progress) :-
    cio_q_d(Cin, QList, Derivable),
    get_sign(Influence, Derivable, Sign),
    !,
    etrace(solve_ir_set, [TrQ, [TrInf], QList, Derivable], resolve),
    etrace(solve_ir_single_res, [TrInf, Sign], resolve),
    etrace(solve_ir_res, [TrQ, Sign], resolve),
    add_resolution_result(Q, TrQ, Sign, Cin, Cnew, Progress, TailProgress,
	                  [UnresolvedIn, Tail], [Unresolved1, NewTail]), % ANA 
    !,
    % note this may fail
    resolve(NewTail, TrTail, Cnew , Cout, Unresolved1, UnresolvedOut, 
            UTracer, UTracerOut, TailProgress).


% adder complete -> construct balance, evaluate & add results

resolve([adder(Q, AddList)|Tail], [adder(TrQ, TrList)|TrTail], Cin, Cout, 
        UnresolvedIn, UnresolvedOut, UTracer, UTracerOut, Progress) :-
    cio_q_d(Cin, QList, Derivable),
    % check quantities in Addlist for known values, transform pos, neg to up, down & noforce
    complete_adder(AddList, Derivable, CompleteAddList, Incomplete, _, _),
    Incomplete = complete,
    etrace(solve_ir_set, [TrQ, TrList, QList, Derivable], resolve),
    etrace(solve_ir_bal, [TrList, Pos, Neg], resolve),
    % make_balance [BalanceValue, min1, min2, ...] = [pos1, pos2, ...] (intern representation)
    make_balance(Q, CompleteAddList, Cin, Ctest1, Balance, BV),
    % add and evaluate balance relation in testcio (balance not added to relations)
    % try_balance gives feedback if append_relation fails
    try_balance(append_relation(Balance, Ctest1, Ctest2, _, false)), % Analyse simple & zero equality should be false    not to create an invalid Q pointers to testcio indexes
    % get value of balancequantity
    cio_d(Ctest2, Derivable2),
    %test quantity sign (also >= and =<)
    get_balance_result(BV, Derivable2, Sign),%because no analyse zero & simple equality, BV is still a valid pointer
    !, %no backtracking on this part
    
    (
    memberchk(Sign, [zero, pos, neg]) % definitive result
    ->
        (
	etrace(solve_ir_bal_res, [TrQ, Pos, Neg, Sign], resolve),
        add_resolution_result(Q, TrQ,Sign, Cin, Cnew, Progress, TailProgress,
	                       [UnresolvedIn, Tail], 
	                       [NewUnresolved, NewTail]),
	!)
    ;   % else: ambiguous result, Sign is a list of possibilities in this case
	(   
	etrace(solve_ir_bal_unres, [TrQ, Pos, Neg, Sign], resolve),
        %add_resolution_result(Q, ambiguous, Cin, Cnew, Progress, TailProgress,
	add_resolution_result(Q, TrQ, Sign, Cin, Cnew, Progress, TailProgress,
	                       [UnresolvedIn, Tail], 
	                       [NewUnresolved, NewTail])
	)
    ),   
    % note this may fail
    resolve(NewTail, TrTail, Cnew , Cout, NewUnresolved, UnresolvedOut, 
	        UTracer, UTracerOut, TailProgress).


% Balance should always be appendible, because BalanceValue is a newly generated quantity
try_balance(X):-
    call(X),!.
try_balance(_):- 
    etrace(solve_append_bal_error, _, general),
    fail.

% appending a result is not always possible, sometimes a derivative is already
% known through another information source (defined in model, inequality etc.)
% In this case the tracer shows that resolution result was not added.
try_append(X):-
    call(X),
    !.
    
try_append(X):- 
    etrace(solve_add_ir_res_fail, X, resolve),
    fail.

% add_resolution_result(+Q, +TrQ, +Sign, +Cin, -Cout, -Progress, +TailProgress,
%	                  +[UnresolvedIn, AdderTailIn], -[UnresolvedOut, AdderTailOut])
% Q is the quantity in question (intern). TrQ is its normal name, for tracing
% Sign is a single value, or a list of values (ambiguity)
% unresolved and adderlist are passed on to update their pointers, 
% because analyse zero / simple equality can change these

% valid new result: let progress be true if not already derivable, 

add_resolution_result(Q, _TrQ, Sign, Cin, Cout, Progress, TailProgress,
	                  [UnresolvedIn, AdderTailIn], [UnresolvedOut, AdderTailOut]):- % ANA 
    memberchk(Sign, [zero, pos, neg]),
	zero_relation_sign(Rel, Sign),
	list_map([], Empty),
	right_order(relation(Q, Rel, Empty), Relation),
    append_split(UnresolvedIn, AdderTailIn, I, OtherAddersIn), % makes one list, for easy passing on with append relatoin. pointers can be updated, but list stays intact.
    try_append(append_relation(Relation, Cin, Cout, Added, true, OtherAddersIn, OtherAddersOut)),
    split_append(UnresolvedOut, AdderTailOut, I, OtherAddersOut), % splits the list again at the same point.
	(Added = [],
	  Progress=TailProgress;
	  Progress=true),
	!.

% ambiguous influence on Q: return all possibilities (upon backtracking)

add_resolution_result(Q, TrQ, Sign, Cin, Cout, Progress, TailProgress,
	                  [UnresolvedIn, AdderTailIn], [UnresolvedOut, AdderTailOut]):- % ANA 
	member(S, Sign),
	nth1(N, Sign, S),
	etrace(pathstart, resolve), %tempflo, cat?
	etrace(solve_ir_amb_nth, [TrQ, N, S], resolve), 
	zero_relation_sign(Rel, S),
	list_map([], Empty),
	right_order(relation(Q, Rel, Empty), Relation),
	once((
        append_split(UnresolvedIn, AdderTailIn, I, OtherAddersIn),
        try_append(append_relation(Relation, Cin, Cout, Added, true, OtherAddersIn, OtherAddersOut)),
        split_append(UnresolvedOut, AdderTailOut, I, OtherAddersOut),
	    (Added = [],
	    Progress=TailProgress;
	    Progress=true)
        )).

% place a relation in the right order: internaly only =, > and >= are used (not < and =<)
right_order(relation(L, <, R), relation(R, >, L)) :- !.
right_order(relation(L, =<, R), relation(R, >=, L)) :- !.
right_order(R, R):- !.


% append_split(+L1, +L2, -I, -All)
% append to lists, return the index where the (virtual) separation is

append_split(L1, L2, I, All):-
    length(L1, I),
    append( L1, L2, All).


% split_append(-L1, -L2, +I, +All) % ANA
% divide a list into two lists splitting at index I

split_append([], All, 0, All):-!.

split_append([H|L1], L2, I, [H|T]):-
    J is I - 1,
    split_append(L1, L2, J, T).


% FL Feb 2007 moved write_addone clauses to interface_garp3.pl


% complete_adder: takes an addList and evaluates to zero
% if sign is known then change notation to up/down/noforce
% otherwise return incomplete and semi transformed list.
% unknown quantities are passed on for tracer

complete_adder([], _, [], complete, [], []):-!. %not yet incomplete -> complete

complete_adder([], _, [], incomplete, [], []):-!. %already incomplete

complete_adder([H|T], Derivable, [Result | NewAddList], Incomplete, [_|TT], TraceList):-
    get_balance_sign(H, Derivable, Result),  %returns: up(Q), down(Q), noforce(Q) or fails
    !,
    complete_adder(T, Derivable, NewAddList, Incomplete, TT, TraceList).

% could not get quantity sign -> set incomplete indicator & continue with rest of list.
complete_adder([H|T], Derivable, [H | NewAddList], incomplete, [HT|TT], [HT|TraceTail]):-
    !,
    complete_adder(T, Derivable, NewAddList, incomplete, TT, TraceTail).


% make_balance(+Quantity, +AddList, +Cin, -Cout, -Balance, -BalanceValue_intern)
% make balance expression in intern representation: [min, min, BV] = [pos, pos]

make_balance(Quantity, AddList, Cin, Cout, relation(Left, =, Right), BV):-
    % create a balance value quantity:
    make_abstract_balance_quantity(Quantity, Cin, Cnew, I, BV),
    % get pointer lists for the positive and negative influences/props.
    add_to_balance(AddList, LeftList, RightList),
	append([I], LeftList, ILeftList),
    % d_q see intern.pl
	double_quantities(ILeftList, NLeftList, [], Cnew, Cnew1),
	double_quantities(RightList, NRightList, [], Cnew1, Cout),
    % create bitmaps for both sides of the balance.
	list_map(NLeftList, Left),
	list_map(NRightList, Right).


% make_abstract_balance_quantity(+Quantity, +Cin, -Cout, -I, -BV), 
% adds a balancequantity in the temporary Cio for influence resolution

make_abstract_balance_quantity(Quantity, Cin, Cout, I, BV):-
    % construct a pointer for the quantiy:
    flag(q_cnt, P, P),
    I is P + 1,
    flag(q_cnt, _, I),
    cio_q_nq(Cin, Cout, Q, NQ),
    list_map([I], BV),
    map_list(Quantity, [X]),
    % construct a name for the quantity:
    memberchk(Name/X, Q),
    Name =.. [_, N], 
    atom_concat(d_, N, QName),
    atom_concat(i_r_balance_, QName, BalanceName),
    % update the quantity list:
    NQ = [BalanceName/I|Q].


% add_to_balance2(+Addlist, -Leftlist,  -Rightlist)

add_to_balance([], [], []).

add_to_balance([up(Q)|Tail], Left, [I|Right]):-
    map_list(Q, [I]),
    add_to_balance(Tail, Left, Right).

add_to_balance([down(Q)|Tail], [I|Left], Right):-
    map_list(Q, [I]),
    add_to_balance(Tail, Left, Right).

add_to_balance([noforce(_Q)|Tail], Left, Right):-
    add_to_balance(Tail, Left, Right).



% get_balance_sign(+Influence, +Derivable, -NewInfluence)
% returns: noforce if influencing quantity is zero, otherwise, leaves influence direction intact,
% but changes from pos/neg to up/down to indicate, that influence has a known relation to zero.

get_balance_sign(up(Quantity), _, up(Quantity)):- !.

get_balance_sign(down(Quantity), _, down(Quantity)):- !.

get_balance_sign(noforce(Quantity), _, noforce(Quantity)):- !.

get_balance_sign(pos(Quantity), Relations, Result):-
    get_quantity_sign(Quantity, Relations, Sign),
    !,
    (Sign = zero ->
    Result = noforce(Quantity);
    Result = up(Quantity)
    ).

get_balance_sign(neg(Quantity), Relations, Result):-
    get_quantity_sign(Quantity, Relations, Sign),
    !,
    (Sign = zero ->
    Result = noforce(Quantity);
    Result = down(Quantity)
    ).


% simplify_relation(+Left, +Right, -NLeft, -NRight)
% If two sides of an (in)equality relation in intern representation contain the same quantities,
% then these can be taken out:
% a + b = a + c -> b = c 
% Sometimes leaving zero on one side:
% a + b = b -> a = 0

% simplify list of relations
simplify_relations([], []).

simplify_relations([relation(Left, Rel, Right)|Tail], [relation(NLeft, Rel, NRight)|NewTail]):-
    simplify_relation(Left, Right, NLeft, NRight),!,
    simplify_relations(Tail, NewTail).

% simplify one relation
simplify_relation(Left, Right, NLeft, NRight):-
    map_without_intersection(Left, Right, NLeft, NRight),
    !.

simplify_relation(Left, Right, Left, Right).


/*** 
*   FL december-2003:  New GARP 2.0 code ends here, 
*   sign_incomplete & sign_add unused by garp 2.0
***/

sign_incomplete(ambiguous, ambiguous):- ! .
% if already 'ambiguous' then unknown sign makes no difference
sign_incomplete(_, incomplete):- ! .
% we can't establish the correct sign

sign_add(first, S, S). 	% first sign
sign_add(zero, zero, zero).
sign_add(pos, pos, pos).
sign_add(neg, neg, neg).
sign_add(pos, zero, pos).
sign_add(zero, pos, pos).
sign_add(neg, zero, neg).
sign_add(zero, neg, neg).
sign_add(pos, neg, ambiguous).
sign_add(neg, pos, ambiguous).

% sign_add(pos_zero, pos, pos).
% sign_add(pos_zero, zero, pos_zero).
% sign_add(pos_zero, neg, ambiguous).   % ????????
% sign_add(neg_zero, neg, neg).
% sign_add(neg_zero, zero, neg_zero).
% sign_add(neg_zero, pos, ambiguous).   % ????????

% sign_add(pos, pos_zero, pos).
% sign_add(zero, pos_zero,  pos_zero).
% sign_add(neg, pos_zero,  ambiguous).  % ????????
% sign_add(neg, neg_zero,  neg).
% sign_add(zero, neg_zero,  neg_zero).
% sign_add(pos, neg_zero,  ambiguous).  % ????????

/***
*   FL december-2003: get_sign still used in garp 2.0 (by other clauses then sign_add!)
***/

get_sign(pos(Quantity), Relations, Sign):-
	get_quantity_sign(Quantity, Relations, Sign).

get_sign(neg(Quantity), Relations, Sign):-
	get_quantity_sign(Quantity, Relations, QSign),
	inverse_sign(QSign, Sign).

inverse_sign(neg, pos).
inverse_sign(pos, neg).
inverse_sign(zero, zero).

get_quantity_sign(Quantity, Relations, Sign):-
	list_map([], M),
	member(relation(Quantity, Relation, M), Relations),
	zero_relation_sign(Relation, Sign),
	!.

get_quantity_sign(Quantity, Relations, Sign):-
	list_map([], M),
	member(relation(M, Relation, Quantity), Relations),
	inverse(Relation, Inverse),
	zero_relation_sign(Inverse, Sign),
	!.

% zero_relation_sign(R, Sign):
% if a parameter has relation R to 0, return Sign

zero_relation_sign(>, pos).
% zero_relation_sign(>=, pos_zero).
zero_relation_sign(=, zero).
zero_relation_sign(<, neg).
% zero_relation_sign(=<, neg_zero).

% get_balance_result(+Quantity, +Relations, -Sign):-
% New for garp 2.0
% Like get quantity sign, but always succeeds:
% returns a sign, e.g. pos, or a list of possibilities in case of ambiguity
% e.g. [pos,zero] or [pos, zero, neg] 

get_balance_result(Quantity, Relations, Sign):-
	list_map([], M),
	member(relation(Quantity, Relation, M), Relations),
	zero_relation_signs(Relation, Sign),
	!.

get_balance_result(Quantity, Relations, Sign):-
	list_map([], M),
	member(relation(M, Relation, Quantity), Relations),
	inverse(Relation, Inverse),
	zero_relation_signs(Inverse, Sign),
	!.

get_balance_result(_, _, [neg, zero, pos]).
	
	
zero_relation_signs(>, pos).
zero_relation_signs(>=, [zero, pos]).
zero_relation_signs(=, zero).
zero_relation_signs(<, neg).
zero_relation_signs(=<, [neg, zero]).


% replace quantities with their intern representation in a list of adders

adders_quantities([], [], Qio, Qio).

adders_quantities([ adder(Q, L) | T ], [ adder(IQ, IL) | NT ], Qin, Qout):-
	intern_quantities([ pos(Q) | L ], [ pos(IQ) | IL ], Qin, Qnew),
	adders_quantities(T, NT, Qnew, Qout).

intern_quantities([], [], Qio, Qio).

intern_quantities([pos(Quantity)|T], [pos(Intern)|NT], Qin, Qout):-
	memberchk(Quantity/Item, Qin), !,
	list_map([Item], Intern),
	intern_quantities(T, NT, Qin, Qout).

intern_quantities([pos(Quantity)|T], [pos(Intern)|NT], Qin, Qout):-
	flag(q_cnt, P, 0), 		/* increase q_cnt */
	I is P + 1,
	flag(q_cnt, _, I),
	list_map([I], Intern),
	intern_quantities(T, NT, [Quantity/I|Qin], Qout),
	!.
intern_quantities([neg(Quantity)|T], [neg(Intern)|NT], Qin, Qout):-
	member(Quantity/Item, Qin), !,
	list_map([Item], Intern),
	intern_quantities(T, NT, Qin, Qout).

intern_quantities([neg(Quantity)|T], [neg(Intern)|NT], Qin, Qout):-
	flag(q_cnt, P, 0), 		/* increase q_cnt */
	I is P + 1,
	flag(q_cnt, _, I),
	list_map([I], Intern),
	intern_quantities(T, NT, [Quantity/I|Qin], Qout),
	!.

/* --------------------------------------------------------------- */

% analyse results of solve
% if simple equality (between two parameters or a parameter and a constant,
% (not zero))
% replace quantity pointer of one of these with the other in all relations

% don't analyse when resolving influences/proportional relations

% FL: in GARP 2.0 two extra arguments: adders in /out
% these also contain pointers and should also be updated.
% This one list should be left intact during the changes: it actually is two lists. 
% (programmers frivolity not to need 4 extra arguments)


analyse_simple_equality(false, _, Q, Q, N, N, Relations, Relations,
	Derivable, Derivable, C, C, A, A). 

analyse_simple_equality(true, [], Q, Q, N, N, Relations, Relations,
	Derivable, Derivable, C, C, A, A). 

analyse_simple_equality(true, [ relation(Left, =, Right) | Tail ], Q, NQ,
 N, NN, Relations, NewRelations, Derivable, NewDerivable, C, NC, A, NA) :- 
	map_list(Left, [ ILeft ]),
	map_list(Right, [ IRight ]),
	replace_index_relations(ILeft, IRight, N, TN),
	replace_index_relations(ILeft, IRight, Relations, TRelations),
	replace_index_relations(ILeft, IRight, Derivable, TDerivable),
	replace_index_relations(ILeft, IRight, Tail, NTail),
	replace_index_relations_for_correspondence(ILeft, IRight, C, TC),
	% fails if that would produce a 'double quantity'
	% sets flag if contradiction found
	replace_index_in_adders(ILeft, IRight, A, TA), % Note the adder pointers are updated
	!,
	replace_index(ILeft, IRight, Q, TQ),
	analyse_simple_equality(true, NTail, TQ, NQ, TN, NN, TRelations,
		NewRelations, TDerivable, NewDerivable, TC, NC, TA, NA). 

analyse_simple_equality(true, [ _ | Tail ], Q, NQ, N, NN, Relations,
		NewRelations, Derivable, NewDerivable, C, NC, A, NA) :- 
	flag(analyse_simple_found_contradiction, false, false),
    analyse_simple_equality(true, Tail, Q, NQ, N, NN, Relations,
		NewRelations, Derivable, NewDerivable, C, NC, A, NA). 

replace_index(_, _, [], []).

replace_index(I1, I2, [ H/I1 | Tail ], [ H/I2 | NewTail ]) :-
	!,
	replace_index(I1, I2, Tail, NewTail).

replace_index(I1, I2, [ H | Tail ], [ H | NewTail ]) :-
	replace_index(I1, I2, Tail, NewTail).

replace_index_relations_for_correspondence(_, _, [], []).
replace_index_relations_for_correspondence(I1, I2,
	[ correspondence(L1, L2) | T ], [correspondence(NL1, NL2) | NT]):-
	replace_index_relations_cor(I1, I2, L1, NL1),
	replace_index_relations_cor(I1, I2, L2, NL2),
	!,
	replace_index_relations_for_correspondence(I1, I2, T, NT).
replace_index_relations_for_correspondence(I1, I2,
	[ if_correspondence(L1, L2) | T ], [if_correspondence(NL1, NL2) | NT]):-
	replace_index_relations_cor(I1, I2, L1, NL1),
	replace_index_relations_cor(I1, I2, L2, NL2),
	!,
	replace_index_relations_for_correspondence(I1, I2, T, NT).


% replace index relations fails if F & S are in a sum together
% or if an inconsistent relation is found: x > x
replace_index_relations(_, _, [], []).

% fail if F & S are already in a sum together

replace_index_relations(F, S, [ relation(Left, _, Right) | _ ], _) :-
	list_map([F, S], M),
	(	map_intersection(M, Left, M)
		;
		map_intersection(M, Right, M)
	),
	!,
	fail.

replace_index_relations(F, S, [ relation(Left, Rel, Right) | Tail ], Result) :-
	m_replace_two(F, S, Left, Right, NLeft, NRight),
	!, 	% fails if F not in Left OR Right
	use_if_not_evident(relation(NLeft, Rel, NRight), NewTail, Result),
	replace_index_relations(F, S, Tail, NewTail).

replace_index_relations(F, S, [ Relation | Tail ], [ Relation | NewTail ]) :-
	replace_index_relations(F, S, Tail, NewTail).


% FL July 2004 
% correspondences can have contradictory relations which just indicates there not 
% active in this state, therefore no use_if_not_evident is used here.

replace_index_relations_cor(_, _, [], []).

% fail if F & S are already in a sum together

replace_index_relations_cor(F, S, [ relation(Left, _, Right) | _ ], _) :-
	list_map([F, S], M),
	(	map_intersection(M, Left, M)
		;
		map_intersection(M, Right, M)
	),
	!,
	fail.

replace_index_relations_cor(F, S, [ relation(Left, Rel, Right) | Tail ], [relation(NLeft, Rel, NRight)|NewTail]) :-
	m_replace_two(F, S, Left, Right, NLeft, NRight),
	!, 	% fails if F not in Left OR Right
	replace_index_relations_cor(F, S, Tail, NewTail).

replace_index_relations_cor(F, S, [ Relation | Tail ], [ Relation | NewTail ]) :-
	replace_index_relations_cor(F, S, Tail, NewTail).


% Patch FL July 2004 Solve failed to find an obvious contradiction:
% probably all parents used already in this particular case
% form was:
% 2 = 1, 1 > 2
% analyse simple constructs the contradictive relation: 1 > 1
% this fails here to make append relation fail thus providing an extra
% chance to capture inconsistencies

use_if_not_evident(relation(M, >, M), Result, Result):-
    !,
    flag(analyse_simple_found_contradiction, _, true),
    etrace(solve_simple_found_invalid, _, add_relation),
    !,
    fail.

use_if_not_evident(Rel, Result, Result):-
	evident(Rel), !.

use_if_not_evident(Rel, Result, [Rel|Result]).


% suceed for fist, try second or succeed
m_replace_two(First, Second, Org1, Org2, New1, New2) :-
	map_replace(First, Second, Org1, New1),
	(map_replace(First, Second, Org2, New2) ; Org2 = New2),
	!.

% failed for first, try second or fail
m_replace_two(First, Second, Org1, Org2, Org1, New2) :-
	map_replace(First, Second, Org2, New2).


/*** FL december-2003: GARP 2.0 new: replace index for adder lists (influence resolution) ***
***  Note that the structure of the list (length, positioning of items is left intact     ***/

replace_index_in_adders(_, _, [], []).

replace_index_in_adders(ILeft, IRight, [Adder|Tail], [NAdder|NTail]):-
    replace_index_single_adder(ILeft, IRight, Adder, NAdder),
    replace_index_in_adders(ILeft, IRight, Tail, NTail).


replace_index_single_adder(ILeft, IRight, adder(Q, Addlist), adder(NQ, NAddlist)):-
    map_replace(ILeft, IRight, Q, NQ),
    !,
    replace_index_addlist(ILeft, IRight, Addlist, NAddlist).

% map replace failed for Q
replace_index_single_adder(ILeft, IRight, adder(Q, Addlist), adder(Q, NAddlist)):-
    replace_index_addlist(ILeft, IRight, Addlist, NAddlist).


replace_index_addlist(_, _, [], []).

replace_index_addlist(ILeft, IRight, [H|T], [NH|NT]):-
    H =.. [Type, Q],
    map_replace(ILeft, IRight, Q, NQ),
    !,
    NH =.. [Type, NQ],
    replace_index_addlist(ILeft, IRight, T, NT).

% map replace failed for Q
replace_index_addlist(ILeft, IRight, [H|T], [H|NT]):-
    replace_index_addlist(ILeft, IRight, T, NT).

/*** FL december-2003: END NEW ***/


/*----------------- analyse zero equality -------------------------*/    

/* FL januari 2004 : analyse_zero_equality  UPDATE FL mar 07: bitversion
*
* in Analyse-simple-equality, y is replaced by x if x=y,
* to minimize the set of relations even further this is also done if:
* q = zero & p = zero because p = q is now derivable in theory.
*
* To do this efficiently all pointers to quantities equal to zero are put in a
* list. Then all these pointers are replaced by the pointer 0, 
* Note that pointers start counting at 1 so 0 is always free.
* A lot of empty relations will come up like:  0 = 0, etc. these are removed 
* Some sums will contain a zero: x + 0 > y, these extra zeros are removed: x > y
* Then zero = 0 is added to the relationsset.
* 
* quantities are equal to zero if: q = zero, or if q >= zero & zero >= q,
* 
* FL march 2004: The latter being unnecessary now that solve can deduct q=zero
* in this case.
*
* Note that zero is mapped to the empty list: [].
*
* When applying the zeroset (changing pointers) relations are checked 
* on contradictions also. eg. 0>0, In that case the operation fails. (message displayed)
*
* This operation is done exhaustively on all known relations, not only on new relations,
* arguments: 
* - the list with pointers equal to zero is easily obtained
* - some call's do not analyse zero equality, but the smaller the relationset, 
*   the faster solve works.
* - double quantity equalities will not be affected.
*
* the operation is only performed when new relations contain a zero equality
* this lowers the number of times this operation is carried out & saves time.
* FL mar 07: NB it is done now always unless no relation in any set has a zero equality
* 
* In Adders and Correspondences indexes are replaced also, 
* empty relations are not removed here. In correspondences they can do no harm.
*
* 
*
* List notation is used, c-predicates for map operations being inadequate:
* Todo: analyse-zero-equality using map operations and new c predicates.
* Not neccesary however for time complexity is not a problem in practice
* 
*/

% analyse simple equality set to false: return input

analyse_zero_equality(false, New, New, Derivable, Derivable, 
                      Base, Base, QList, QList, Corr, Corr, Adders, Adders).

% analyse zero equality set to false in switch: return input

analyse_zero_equality(_, New, New, Derivable, Derivable, 
                      Base, Base, QList, QList, Corr, Corr, Adders, Adders):-
                      flag(no_analyse_zero, Flag, Flag),
                      algorithm_assumption_flag(Flag, true),
                      !.

/*FL 22-2-2007 do it always, bug found... 
NEW mar 07 , don't do it when no equality to zero present what so ever
New bitvector version
*/
% relations do not contain a zero equality: return input

analyse_zero_equality(true, New, New, Derivable, Derivable, 
                      Base, Base, QList, QList, Corr, Corr, Adders, Adders):-     
    % smallest
    % map_list_for_relation_set(New, New1),
    (select(relation(bitvector(1), =, bitvector(0)), New, NewR) ; New = NewR),
    \+ contains_equality_to_zero(NewR), 
    % next
    % map_list_for_relation_set(Base, Base1),
    (select(relation(bitvector(1), =, bitvector(0)), Base, BaseR) ; Base = BaseR),
    \+ contains_equality_to_zero(BaseR), 
    % biggest set last
    % map_list_for_relation_set(Derivable, Derivable1),
    (select(relation(bitvector(1), =, bitvector(0)), Derivable, DerivableR) ; Derivable = DerivableR),
    \+ contains_equality_to_zero(DerivableR), 
    !.


% analyse zero equality bitvector version

analyse_zero_equality(true, NewIn, NewOut, DerivableIn, DerivableOut, 
                      BaseIn, BaseOut, QListIn, QListOut, CorrIn, CorrOut, AddersIn, AddersOut):-     
    % find all zero indexes
     % flstop,
%    map_list_for_relation_set(NewIn, New),
%    map_list_for_relation_set(DerivableIn, Derivable),
%    map_list_for_relation_set(BaseIn, Base),
%    map_list_for_correspondences(CorrIn, Corr),
%    map_list_for_adders(AddersIn, Adders),
%    get_zero_set(Derivable, ZerosL),
%    delete(ZerosL, 0, ZerosetL), % after the first zero-analysis, 0 is always part of the list
			
    get_bitwise_zero_set(DerivableIn, Zeros, BitZeros),
    
 %   (Zeros = ZerosL ; flstop),
    
    
    delete(Zeros, 0, Zeroset), % after the first zero-analysis, 0 is always part of the list
    % delete bitwise 0 (NB bitmap 1 corresponds to [0])
    NotZero is \1,
    BitZeroset is BitZeros /\ NotZero,
    InvBitZeroset is \BitZeros,
    
%     (Zeroset = ZerosetL ; flstop),
    
%    apply_zero_set(Derivable, Zeroset, Derivable1),
%    apply_zero_set(New, Zeroset, New1),
%    apply_zero_set(Base, Zeroset, Base1),
%    apply_zero_set_correspondences(Corr, Zeroset, Corr1),
%    apply_zero_set_adders(Adders, Zeroset, Adders1),
        % convert back to map representation
%    list_map_for_relation_set(New1, NewOutL),
%    list_map_for_relation_set(Derivable1, DerivableOutL),
%    list_map_for_relation_set(Base1, BaseOutL),
%    list_map_for_correspondences(Corr1, CorrOutL),
%    list_map_for_adders(Adders1, AddersOutL),

        % change all zero indexes to 0, 
        % remove extra 0's & evident relations, 
        % add [] = [0], 
        % find contradiction?  x>x then fail 
    apply_bitwise_zero_set(DerivableIn, InvBitZeroset, DerivableOut),
    apply_bitwise_zero_set(NewIn, InvBitZeroset, NewOut),
    apply_bitwise_zero_set(BaseIn, InvBitZeroset, BaseOut),
    apply_bitwise_zero_set_correspondences(CorrIn, InvBitZeroset, CorrOut),
    apply_bitwise_zero_set_adders(AddersIn, BitZeroset, InvBitZeroset, AddersOut),
        
 
%    (NewOutL == NewOut ; flstop),
%    ( DerivableOut == DerivableOutL ; flstop),
%    ( BaseOut == BaseOutL ; flstop),
%    ( CorrOut == CorrOutL ; flstop),
%    ( AddersOut == AddersOutL ; flstop),
 
    
    % update Qlist
    replace_indexes_with_zeros(Zeroset, QListIn, QListOut).







/*-------------- Analyse zero equality tools: ----------------------*/

% set of relations has at least one relation to zero. memberchk version with bitvectors...:)

contains_equality_to_zero(List):- 
	memberchk(relation(bitvector(0), =, bitvector(_)), List),!.

contains_equality_to_zero(List):- 
	memberchk(relation(bitvector(_), =, bitvector(0)), List),!.


% get_bitwise_zero_set(+Relations, -Zeroset, -Bitwisezeros) 
% FL mar 07: bitwise version
% find all quantities / pointers equal to zero in a relationset:
% intern-list representation, use q = [] or q>=[] /\ []>=q 
% NOW: q = bitvector(0) etc.  NB [] = bitvector(0), [0] is bitvector(1),

% done
get_bitwise_zero_set([], [], 0).

% X equal to zero, add to zeroset
get_bitwise_zero_set([relation(bitvector(X), =, bitvector(0))|Tail], [Xindex|ZTail], BitZeros):-
	% X is single item: e.g. list: [3], not addition like [3, 5]
	1 is popcount(X),
	!,
	map_list(bitvector(X), [Xindex]),
	get_bitwise_zero_set(Tail, ZTail, BZTail),
	% take the union of tail and X
	BitZeros is BZTail \/ X.

% X equal to zero, add to zeroset
get_bitwise_zero_set([relation(bitvector(0), =, bitvector(X))|Tail], [Xindex|ZTail], BitZeros):-
	% X is single item: e.g. list: [3], not addition like [3, 5]
	1 is popcount(X),
	!,
	map_list(bitvector(X), [Xindex]),
	get_bitwise_zero_set(Tail, ZTail, BZTail),
	% take the union of tail and X
	BitZeros is BZTail \/ X.

% X equal or greater to zero if also inverse then add to zeroset
get_bitwise_zero_set([relation(bitvector(X), >=, bitvector(0))|Tail], [Xindex|ZTail], BitZeros):-
	% X is single item: e.g. list: [3], not addition like [3, 5]
	1 is popcount(X),
	bitwise_combine_to_zero(right, X, Tail),
	!,
	map_list(bitvector(X), [Xindex]),
	get_bitwise_zero_set(Tail, ZTail, BZTail),
	% take the union of tail and X
	BitZeros is BZTail \/ X.


% X equal or greater to zero if also inverse then add to zeroset
get_bitwise_zero_set([relation(bitvector(0), >=, bitvector(X))|Tail], [Xindex|ZTail], BitZeros):-
	% X is single item: e.g. list: [3], not addition like [3, 5]
	1 is popcount(X),
	bitwise_combine_to_zero(left, X, Tail),
	!,
	map_list(bitvector(X), [Xindex]),
	get_bitwise_zero_set(Tail, ZTail, BZTail),
	% take the union of tail and X
	BitZeros is BZTail \/ X.

% no zero equality, next relation
get_bitwise_zero_set([_|Tail], ZTail, BitZeros):-
    get_bitwise_zero_set(Tail, ZTail, BitZeros).


bitwise_combine_to_zero(right, X, Relations):-
    memberchk(relation(bitvector(0), >=, bitvector(X)), Relations).

bitwise_combine_to_zero(left, X, Relations):-
    memberchk(relation(bitvector(X), >=, bitvector(0)), Relations).



% apply zeros for Qlist, use analyse simple equality predicate. NB still List operated!
replace_indexes_with_zeros([], In, In).

replace_indexes_with_zeros([H|T], In, Out):-
    replace_index(H, 0, In, New),
    replace_indexes_with_zeros(T, New, Out).


% apply_zero_set(+RelationsIn, +Zeroset, -RelationsOut)
% apply zeros to set of relations in intern-list representation
% append the relation 0 = zero, which is removed because evident.

apply_bitwise_zero_set(RelationsIn, IZeroset, RelationsOut):-
    apply_bitwise_zeros(RelationsIn, IZeroset, Relations1),    % fails in case of contradiction
    append([relation(bitvector(1), =, bitvector(0))], Relations1, RelationsOut).


% apply_zeros(+RelationsIn, +Zeroset, -RelationsOut)
% apply zeros to set of relations in intern-bit representation,
% fails with error report in case of a contradiction

apply_bitwise_zeros([], _, []).

% valid new relation after zero application, return, continue with rest

apply_bitwise_zeros([relation(X, Rel, Y)|Tail], IZeros, Out):-
    % change indexes in zeroset to zero, leave one zero if no other indexes.
    change_bitwise_indexes_to_one_zero(X, IZeros, X1),
    change_bitwise_indexes_to_one_zero(Y, IZeros, Y1),
    bitwise_check_relation(X1, Rel, Y1, Result), % (0=0 (empty)) or x>x (contradiction) or x=x evident or x?y(normal)
    !,
    determine_b_c_r_result(Result, relation(X1, Rel, Y1), NewTail, Out),
    apply_bitwise_zeros(Tail, IZeros, NewTail).

% keep relation
determine_b_c_r_result(normal, relation(X, Rel, Y), NewTail, [relation(X, Rel, Y)|NewTail]).

% discard relation
determine_b_c_r_result(evident, _, Out, Out).

% fail on contradiction
determine_b_c_r_result(contradiction, _, _, _):-
	etrace(solve_zero_found_invalid, _, add_relation),
	!, 
	fail.

% change_indexes_to_one_zero(+BitmapIn, +Zerosmap, -BitmapOut)
% remove all elements in zeros from list, leave one zero if no other 
% element/pointer is present 
% eg. [4, 5], zeros=5 -> [4]
% eg. [5], zeros=5 -> [0]
% eg. [], zeros=5 -> []
change_bitwise_indexes_to_one_zero(bitvector(0), _InvZeromap, bitvector(0)):-
	!.
change_bitwise_indexes_to_one_zero(bitvector(In), InvZeromap, bitvector(Out)):-
	Out1 is In /\ InvZeromap, % subtract the zeroset
	(   
	  Out1 == 0 % []
	->  
	  Out = 1  %-> [0]
	;   
	  Out1 = Out % -> [X, Y, Z]
	).


% check relation for information value

bitwise_check_relation(X, Rel, X, Result):- 
    !,
    (Rel == > 
    -> 
    Result = contradiction % x > x is impossible
    ; 
    Result = evident).  % x = x or x >= x is evident.

bitwise_check_relation(X, Rel, Y, Result):-
    %list is empty or contains only 0's
    only_bitwise_zeros_in_list(X),
    only_bitwise_zeros_in_list(Y),
    !,
    (Rel == > 
    -> 
    Result = contradiction  % zero > zero is impossible
    ; 
    Result = evident).    % zero = zero is evident empty/no information value

% case else: normal
bitwise_check_relation(_X, _Rel, _Y, normal).
	


% succeed if list is empty or contains only 0's
% bitwise check is very easy... :)
only_bitwise_zeros_in_list(bitvector(0)).

only_bitwise_zeros_in_list(bitvector(1)).


/* apply for correspondences & adders */


% apply_zero_set_correspondences(+CorrespondencesIn, +Zeroset, -CorrespondencesOut)
% Pointers in correspondences (intern-list representation) need to be updated.
% all relations are returned:
% an if_correspondence([relation( 0 > 0), ...) will not fire 
% an if_correspondence([relation( 0 = 0), ...) will fire 
% so information is preserved, however, correspondences containing no
% information are formed. They can do no harm though.

apply_bitwise_zero_set_correspondences([], _IZeroset, []).

apply_bitwise_zero_set_correspondences([H|T], IZeroset, [NH|NT]):-
    H =.. [C, Relations1, Relations2],
    put_all_bitwise_zero_indexes(Relations1, IZeroset, NR1),
    put_all_bitwise_zero_indexes(Relations2, IZeroset, NR2),
    NH =.. [C, NR1, NR2],
    apply_bitwise_zero_set_correspondences(T, IZeroset, NT).


% apply_zero_set_adders(+AddersIn, +Zeroset, +AddersOut)
% Pointers in adders (intern-list representation) need to be updated also.

apply_bitwise_zero_set_adders([], _Zeroset, _InvZeroset, []).

apply_bitwise_zero_set_adders([H|T], Zeroset, InvZeroset, [NH|NT]):-
    H =.. [adder, Q, List],
    replace_bitwise_zero_indexes(Q, InvZeroset, NQ),
    put_all_bitwise_zero_indexes_addlist(List, Zeroset, NList),
    NH =.. [adder, NQ, NList],
    apply_bitwise_zero_set_adders(T, Zeroset, InvZeroset, NT).


%replace all zero indexes with 0 in a relation. 
put_all_bitwise_zero_indexes([], _, []).

put_all_bitwise_zero_indexes([relation(X, Rel, Y)|Tail], IZList,[relation(X1, Rel, Y1)|NTail]):-
    replace_bitwise_zero_indexes(X, IZList, X1),
    replace_bitwise_zero_indexes(Y, IZList, Y1),
    !,
    put_all_bitwise_zero_indexes(Tail, IZList, NTail).
    
% do not replace empty: [] with zero [0]
replace_bitwise_zero_indexes(bitvector(0), _Zeromap, bitvector(0)):-
	!.
replace_bitwise_zero_indexes(bitvector(In), InvZeromap, bitvector(Out)):-
	%InverseZeros is \Zeromap,
	Out1 is In /\ InvZeromap, %subtract the zeroset.
	(   
	  Out1 == 0 % []
	->  
	  Out = 1  %-> [0]
	;   
	  Out1 = Out % -> [X, Y, Z]
	).
	

% replace all zero indexes with 0 in an addlist.
% eg. [pos(X), neg(Y)]
 
put_all_bitwise_zero_indexes_addlist([], _, []).

put_all_bitwise_zero_indexes_addlist([H|T], ZMap,[NH|NT]):-
    H =.. [X, bitvector(Q)],
    Res is Q /\ ZMap,
    (	
      Res == 0 % Q is not in Zmap
    ->	
      NQ = Q
    ;	
      NQ = 1  % bitvector 1 is the zero pointer: [0]
    ),
    !,
    NH =.. [X, bitvector(NQ)], % X could be: noforce, saves evaluation later on...
    put_all_bitwise_zero_indexes_addlist(T, ZMap, NT).






/*** FL januari 2004: NEW GARP 2.0 SPECIFIC ENDS HERE ***/



/* correspondence relations
   inspected when a correspondence relation is found, or
   when a new relation is derived
 */

/* FL 7-3-2004: Old version: has a...
* Bug: when analyse simple encounters an x=y relation, it replaces all y's with
* x and removes the x=y (now x=x) relation. If this x=y relation is the condition part of
* a correspondence, this results in an empty list. The empty list should mean:
* no conditions -> fire correspondence. But the empty list fails the is_derivable
* check. (common_select fails on the empty list).
* an empty list as consequence is of no concern; append_relation_all is a list
* operation and has no problem with empty lists.

inspect_correspondence(Relations, Cin, Cout):-
	inspect_correspondence(Relations, Cin, Cout, []).

inspect_correspondence(Relations, Cin, Cout, Previous):-
	cio_c_nc_d(Cin, Cnew, Correspondence, RCorrespondence, Derivable), 
(	common_select(Correspondence, 
		correspondence(L1, L2), 
		RCorrespondence)
;	common_select(Correspondence, 
		correspondence(L2, L1), 
		RCorrespondence)
;	common_select(	Correspondence, 
		if_correspondence(L1, L2), 
		RCorrespondence)
), 
	common_select(L1, One, RL1), 
	is_derivable(One, Relations), 
	check_derivable(RL1, Derivable), 
	append(Previous, L2, ToAdd), 
	!, 
	inspect_correspondence(Relations, Cnew, Cout, ToAdd).

inspect_correspondence(_, Cio, Cio, []):- !.
inspect_correspondence(_, Cin, Cout, ToAdd):-
	append_relation_all(ToAdd, Cin, Cout, _, true).
	
*** New version below: ***/

inspect_correspondence(Relations, Cin, Cout, DoAnalyseSimple):-
	inspect_correspondence(Relations, Cin, Cout, [], DoAnalyseSimple).

inspect_correspondence(Relations, Cin, Cout, Previous, DoAnalyseSimple):-
	cio_c_nc_d(Cin, Cnew, Correspondence, RCorrespondence, Derivable),
(	common_select(Correspondence,
		correspondence(L1, L2),
		RCorrespondence)
;	common_select(Correspondence,
		correspondence(L2, L1),
		RCorrespondence)
;	common_select(	Correspondence,
		if_correspondence(L1, L2),
		RCorrespondence)
),
	(
	common_select(L1, One, RL1),
	is_derivable(One, Relations),
	check_derivable(RL1, Derivable)
	;       % new:
	L1 = [] % An empty list means no conditions: fire correspondence.
	),
	append(Previous, L2, ToAdd),
	!,
	inspect_correspondence(Relations, Cnew, Cout, ToAdd, DoAnalyseSimple).

inspect_correspondence(_, Cio, Cio, [], _):- !.
inspect_correspondence(_, Cin, Cout, ToAdd, DoAnalyseSimple):-
	append_relation_all(ToAdd, Cin, Cout, _, DoAnalyseSimple).

inspect_a_correspondence(correspondence(L1, L2), Cin, Cout):-
	cio_d(Cin, Derivable),
	check_derivable(L1, Derivable),
	!,
	append_relation_all(L2, Cin, Cout, _, true).

inspect_a_correspondence(correspondence(L1, L2), Cin, Cout):-
	cio_d(Cin, Derivable),
	check_derivable(L2, Derivable),
	!,
	append_relation_all(L1, Cin, Cout, _, true).

inspect_a_correspondence(if_correspondence(L1, L2), Cin, Cout):-
	cio_d(Cin, Derivable),
	check_derivable(L1, Derivable),
	!,
	append_relation_all(L2, Cin, Cout, _, true).

inspect_a_correspondence(_, Cio, Cio).




/*************** old listversion analyse zero preds..
%%	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	
%	
%	
%	
%	
%	



% get_zero_set(+Relations, -Zeroset)
% find all quantities / pointers equal to zero in a relationset:
% intern-list representation, use q = [] or q>=[] /\ []>=q

% done
get_zero_set([], []).

% X equal to zero, add to zeroset
get_zero_set([relation([X], =, [])|Tail], [X|ZTail]):-
    !,
    get_zero_set(Tail, ZTail).

% X equal to zero, add to zeroset
get_zero_set([relation([], =, [X])|Tail], [X|ZTail]):-
    !,
    get_zero_set(Tail, ZTail).

% X equal or greater to zero if also inverse then add to zeroset
get_zero_set([relation([X], >=, [])|Tail], [X|ZTail]):-
    combine_to_zero(right, X, Tail),
    !,
    get_zero_set(Tail, ZTail).

% X equal or greater to zero if also inverse then add to zeroset
get_zero_set([relation([], >=, [X])|Tail], [X|ZTail]):-
    combine_to_zero(left, X, Tail),
    !,
    get_zero_set(Tail, ZTail).

% no zero equality, next relation
get_zero_set([_|Tail], ZTail):-
    get_zero_set(Tail, ZTail).


combine_to_zero(right, X, Relations):-
    memberchk(relation([], >=, [X]), Relations).

combine_to_zero(left, X, Relations):-
    memberchk(relation([X], >=, []), Relations).
% apply_zero_set(+RelationsIn, +Zeroset, -RelationsOut)
% apply zeros to set of relations in intern-list representation
% append the relation 0 = zero, which is removed because evident.

apply_zero_set(RelationsIn, Zeroset, RelationsOut):-
    apply_zeros(RelationsIn, Zeroset, Relations1),    % fails in case of contradiction
    append([relation([0], =, [])], Relations1, RelationsOut).


% apply_zeros(+RelationsIn, +Zeroset, -RelationsOut)
% apply zeros to set of relations in intern-list representation,
% fails with error report in case of a contradiction

apply_zeros([], _, []).

% valid new relation after zero application, return, continue with rest

apply_zeros([relation(X, Rel, Y)|Tail], Zeros, [relation(X1, Rel, Y1)|NewTail]):-
    % change indexes in zeroset to zero, leave one zero if no other indexes.
    change_indexes_to_one_zero(X, Zeros, X1, false),
    change_indexes_to_one_zero(Y, Zeros, Y1, false),
    \+ empty_relation(X1, Rel, Y1, _Result), 
    !,
    %\+ invalid_relation(relation(X1, Rel, Y1)), % X > X (contradiction) fail (FL July 2004: superfluous empty relation checks for contradiction)
    %!,
    apply_zeros(Tail, Zeros, NewTail).

% empty relation after zero application, discard, continue with rest

apply_zeros([relation(X, Rel, Y)|Tail], Zeros, NewTail):-
    % change indexes in zeroset to zero, leave one zero if no other indexes.
    change_indexes_to_one_zero(X, Zeros, X1, false),
    change_indexes_to_one_zero(Y, Zeros, Y1, false),
    empty_relation(X1, Rel, Y1, Result),% 0 = 0 (empty) or x>x (contradiction) or x=x evident
    memberchk(Result, [empty, evident]),
    !,
    apply_zeros(Tail, Zeros, NewTail).


% change_indexes_to_one_zero(+ListIn, +Zeros, -ListOut, +false)
% remove all elements in zeros from list, leave one zero if no other 
% element/pointer is present (indicater set to true).
% eg. [4, 5], zeros=5 -> [4]
% eg. [5], zeros=5 -> [0]

% done
change_indexes_to_one_zero([], _, [], _). 

% no other pointers unequal to zero return [0]
change_indexes_to_one_zero([X], Zeros, [0], false):-
    memberchk(X, Zeros),!.

% other pointers found before, return empty
change_indexes_to_one_zero([X], Zeros, [], true):-
    memberchk(X, Zeros),!.

% nonzero pointer, return it.
change_indexes_to_one_zero([X], _, [X], _).

% zero pointer in front, remove from list.
change_indexes_to_one_zero([X, Y|Tail], Zeros, NewTail, NonZeroInList):-
    memberchk(X, Zeros),!,
    change_indexes_to_one_zero([Y|Tail], Zeros, NewTail, NonZeroInList).

% nonzero pointer, return it, set indicator, to true
change_indexes_to_one_zero([X, Y|Tail], Zeros, [X|NewTail], _NonZeroInList):-
    change_indexes_to_one_zero([Y|Tail], Zeros, NewTail, true).




% check relation for information value

empty_relation(X, Rel, X, Result):- 
    !,
    (Rel == > 
    -> 
    Result = contradiction, % x > x is impossible
    etrace(solve_zero_found_invalid, _, add_relation)
    ; 
    Result = evident).  % x = x or x >= x is evident.

empty_relation(X, Rel, Y, Result):-
    %list is empty or contains only 0's
    only_zeros_in_list(X),
    only_zeros_in_list(Y),
    !,
    (Rel == > 
    -> 
    Result = contradiction,  % zero > zero is impossible
    etrace(solve_zero_found_invalid, _, add_relation)
    ; 
    Result = empty).    % zero = zero is evident empty/no information value




% succeed if list is empty or contains only 0's
only_zeros_in_list([]).

only_zeros_in_list([0|T]):-
    only_zeros_in_list(T).


% * apply for correspondences & adders *


% apply_zero_set_correspondences(+CorrespondencesIn, +Zeroset, -CorrespondencesOut)
% Pointers in correspondences (intern-list representation) need to be updated.
% all relations are returned:
% an if_correspondence([relation( 0 > 0), ...) will not fire 
% an if_correspondence([relation( 0 = 0), ...) will fire 
% so information is preserved, however, correspondences containing no
% information are formed. They can do no harm though.

apply_zero_set_correspondences([], _Zeroset, []).

apply_zero_set_correspondences([H|T], Zeroset, [NH|NT]):-
    H =.. [C, Relations1, Relations2],
    put_all_zero_indexes(Relations1, Zeroset, NR1),
    put_all_zero_indexes(Relations2, Zeroset, NR2),
    NH =.. [C, NR1, NR2],
    apply_zero_set_correspondences(T, Zeroset, NT).


% apply_zero_set_adders(+AddersIn, +Zeroset, +AddersOut)
% Pointers in adders (intern-list representation) need to be updated also.

apply_zero_set_adders([], _Zeroset, []).

apply_zero_set_adders([H|T], Zeroset, [NH|NT]):-
    H =.. [adder, Q, List],
    replace_zero_indexes(Q, Zeroset, NQ),
    put_all_zero_indexes_addlist(List, Zeroset, NList),
    NH =.. [adder, NQ, NList],
    apply_zero_set_adders(T, Zeroset, NT).


%replace all zero indexes with 0 in a relation. 
put_all_zero_indexes([], _, []).

put_all_zero_indexes([relation(X, Rel, Y)|Tail], ZList,[relation(X1, Rel, Y1)|NTail]):-
    replace_zero_indexes(X, ZList, X1),
    replace_zero_indexes(Y, ZList, Y1),
    !,
    put_all_zero_indexes(Tail, ZList, NTail).
    
replace_zero_indexes([], _, []).

replace_zero_indexes([I|T], ZList, [0|NT]):-
    memberchk(I, ZList),
    !,
    replace_zero_indexes(T, ZList, NT).

replace_zero_indexes([H|T], ZList, [H|NT]):-
    replace_zero_indexes(T, ZList, NT).


% replace all zero indexes with 0 in an addlist.
% eg. [pos(X), neg(Y)]
 
put_all_zero_indexes_addlist([], _, []).

put_all_zero_indexes_addlist([H|T], ZList,[NH|NT]):-
    H =.. [X, [Q]],
    memberchk(Q, ZList),
    !,
    NH =.. [X, [0]], % X could be: noforce, saves evaluation later on...
    put_all_zero_indexes_addlist(T, ZList, NT).

put_all_zero_indexes_addlist([H|T], ZList,[H|NT]):-
    put_all_zero_indexes_addlist(T, ZList, NT).


%*--------------special list to map and back tools for sets--------*%

% list_map_for_relation_set(+InternRelationsIn, -InternListRelationsOut)
% From LIST to Map
% convert bitmap representation to list (map_list), for a whole
% set of relations (intern representation), 
% making them intern-list representation

list_map_for_relation_set([], []).

list_map_for_relation_set([relation(X, Rel, Y)|Tail], [relation(X1, Rel, Y1)|NewTail]):-
    list_map(X, X1),
    list_map(Y, Y1),
    list_map_for_relation_set(Tail, NewTail).


% map_list_for_relation_set(+InternListRelationsOut, -InternRelationsIn)
% From MAP to LIST
% convert list representation to bitmap (list_map), for a whole
% set of relations (intern-list representation), 
% making them intern representation

map_list_for_relation_set([], []).

map_list_for_relation_set([relation(X, Rel, Y)|Tail], [relation(X1, Rel, Y1)|NewTail]):-
    map_list(X, X1),
    map_list(Y, Y1),
    map_list_for_relation_set(Tail, NewTail).


% From LIST to Map for correspondences

list_map_for_correspondences([], []).

list_map_for_correspondences([H|T], [NH|NT]):-
    H =.. [C, Relations1, Relations2],
    list_map_for_relation_set(Relations1, NR1),
    list_map_for_relation_set(Relations2, NR2),
    NH =.. [C, NR1, NR2],
    list_map_for_correspondences(T, NT).


% From LIST to Map for adders

list_map_for_adders([], []).

list_map_for_adders([H|T], [NH|NT]):-
    H =.. [adder, Q, List],
    list_map(Q, NQ),
    list_map_addlist(List, NList),
    NH =.. [adder, NQ, NList],
    list_map_for_adders(T, NT).

list_map_addlist([], []).

list_map_addlist([H|T], [NH|NT]):-
    H =.. [X, Q],
    list_map(Q, NQ),
    NH =.. [X, NQ],
    list_map_addlist(T, NT).


% From MAP to LIST for correspondences

map_list_for_correspondences([], []).

map_list_for_correspondences([H|T], [NH|NT]):-
    H =.. [C, Relations1, Relations2],
    map_list_for_relation_set(Relations1, NR1),
    map_list_for_relation_set(Relations2, NR2),
    NH =.. [C, NR1, NR2],
    map_list_for_correspondences(T, NT).


% From MAP to LIST for adders

map_list_for_adders([], []).

map_list_for_adders([H|T], [NH|NT]):-
    H =.. [adder, Q, List],
    map_list(Q, NQ),
    map_list_addlist(List, NList),
    NH =.. [adder, NQ, NList],
    map_list_for_adders(T, NT).

map_list_addlist([], []).

map_list_addlist([H|T], [NH|NT]):-
    H =.. [X, Q],
    map_list(Q, NQ),
    NH =.. [X, NQ],
    map_list_addlist(T, NT).

%*-------- End listmap maplist tools ---------*%


% End listversion analyse zero preds...................*/
