/*
versionpatch.pl

Version patch code. Used by garpModel->versionPatch when a model is loaded that uses an older modelDefinitionVersion than the current one.

Loaded only when needed.

This file defined pl_doVersionPatch(Model, CurrentVersion, ModelVersion) to get Model from ModelVersion to CurrentVersion.
For every step (from version 0 to 1, 1 to 2 etc) is has code, so we can patch any old version.

8-12-2005 Jelmer Jellema (Spin in het Web)
Part of GARP3
See copyright notice

*/

%
pl_doVersionPatch(M,CurrentVersion,LowerVersion):-
	%gp3
	%check if there is a patch for models being at a lower definition level
	
	LowerVersion < CurrentVersion,!,
	ignore(pl_versionPatch(M,LowerVersion)),
	NextVersion is LowerVersion + 1,
	pl_doVersionPatch(M,CurrentVersion,NextVersion),
	send(@model, changed, @on).
%
pl_doVersionPatch(_,_,_). %done
%%
/*
pl_versionPatch(M,OldVersion)
	gp3 0.3 rewrite of old versionpatch
	this one is called for every version level between the current one
	and the model's one
	pl_versionPatch(M,1) means: patch M so that it current version (version 1) goes to the next one (version 2)
	so we can implement patches for any version wanted
	Version 0 is the code we used to patch stuff before explicit versioning
	So here we have to do checks about the content
*/

pl_versionPatch(M,14):-
    % This missing Sketches bug should NEVER EVER reappear anymore
    (
	pl_checkMissingVersionPatch(M,8)
    ;
	true
    ),
    @pce->>write_ln('14 --> 15: Update done.').


pl_versionPatch(M,13):-
    (
	pl_checkMissingVersionPatch(M,8) % some models missed this patch
    ;    
       % add Sketch definitions for quantities in stgraph 
	@pce->>write_ln('13 --> 14: Add quantity definitions to sketch state-transition graph'),
        get(M, hypered, stgraphMF, SK),
        get(SK?elements, find_all, message(@arg1, instance_of, sketchStateElement), 
	    SketchStates), 
        new(AllIneqs, chain), 
	SketchStates->>for_all(
	    and(
		message(AllIneqs, union, @arg1?sortedSketchStateInequalities)
	    )
	),    
        pl_extract_quantities(AllIneqs, SketchQuantities),
	send(SketchQuantities, sort, ?(@arg1?name, compare, @arg2?name)),
	send(SK, sketchQuantities, SketchQuantities),
        @pce->>write_ln('13 --> 14: Update done.')
    ).


pl_versionPatch(M,12):-
    (
	pl_checkMissingVersionPatch(M,8) % some models missed this patch
    ;    
	/* Open an editor for each model fragment and close it again (to save the relative layout) */
	(
	    @app<<-hypered(pleaseWait) ->
	    true
	;
	    W *= pleaseWait('Please wait while your model is being updated.\nThis will involve some flashing screens.\nPlease save your model afterwards.'),
	    @app->>hyper(W,pleaseWait)
        ),
	get(M, modelFragments, MFsAndScenarios),
	chain_list(MFsAndScenarios, MFsAndScenariosList),
	forall(
	    member(MForScenario, MFsAndScenariosList),
	    (
		send(@app, openViewEditor, MForScenario),
		get(M, all_hypers, HyperChain),
		chain_list(HyperChain, HyperList),
		forall(
		    member(Hyper, HyperList),
		    (
			get(Hyper, to, Editor),
			get(Editor, class_name, viewEditor) ->
			send(Editor, destroy)
		    ;
			true
		    )
		)
	    )
	),  
	send(@app, thankYou),

        % previous patch (11) could leave StructRelDefs empty, which was not intended
        % if this is the case, initialise them again
        get(M, hypered, structureMF, SM),
        get(SM, sketchStructuralRelationDefinitions, StructRelDefs), 
        if
           StructRelDefs->>empty
        then
        (
           send(SM, init_structural_relation_definitions), 
           @pce->>write_ln('12 --> 13: Updated definitions for sketch structural relations in structure model.')
        ), 

	@pce->>write_ln('12 --> 13: Update done.')
    ).

pl_versionPatch(M,11):-
    (
	pl_checkMissingVersionPatch(M,8) % some models missed this patch
    ;    
       % update Sketch definitions for sketch relations in concept map and structure model
       @pce->>write_ln('11 --> 12: Updated definitions for sketch relations in concept map and structure model.'),

       % get relation definitions and move them to the sketch to which they belong
       % for the concept map
       get(M, sketchRelationDefinitions, RelDefs),
       get(M, hypered, conceptMapMF, CM), 
       send(CM, slot, sketchRelationDefinitions, RelDefs), 
       % the old slot should not be used anymore
       send(M, slot, sketchRelationDefinitions, new(chain)), 

       % get relation definitions and move them to the sketch to which they belong
       % for the structural model
       get(M, sketchStructuralRelationDefinitions, StructRelDefs),
       get(M, hypered, structureMF, SM),
       send(SM, slot, sketchStructuralRelationDefinitions, StructRelDefs), 
       % the old slot should not be used anymore
       send(M, slot, sketchStructuralRelationDefinitions, new(chain)), 

       M->>savedSketches(new(hash_table)),
       @pce->>write_ln('11 --> 12: Update done.')
    ).

pl_versionPatch(M,10):-
    (
	pl_checkMissingVersionPatch(M,8) % some models missed this patch
    ;
       % update Sketch definitions for Processes, Agents and Scenarios
       @pce->>write_ln('10 --> 11: Updated definitions for sketch processes, agents, and scenarios.'),
       get(M, sketchProcessDefinitions, ProcessDefs), 
       get(M, sketchAgentDefinitions, AgentDefs), 
       get(M, sketchScenarioDefinitions, ScenarioDefs), 
       pl_updateSketchDefinitions(ProcessDefs), 
       pl_updateSketchDefinitions(AgentDefs), 
       pl_updateSketchDefinitions(ScenarioDefs)
    ).

pl_versionPatch(M,9):-
    (
	pl_checkMissingVersionPatch(M,8) % some models missed this patch
    ;
	@pce->>write_ln('9 --> 10: Updated metadata definitions.'),       
	send(M, updateMetaDataFields)
    ).

% for Sketch, AB, Aug 2006
pl_versionPatch(M,8):-
	%from 8 to 9 added garpModel->sketches and related elements, AB, july 2006 + aug 2006
	% Added missing metadata fields
	
	@pce->>write_ln('8 --> 9: Added sketch functionality and metadata fields.'),
	M->>slot(sketchScenarioDefinitions,new(chain)),
	M->>slot(sketchAgentDefinitions,new(chain)),
	M->>slot(sketchProcessDefinitions,new(chain)),
	M->>slot(sketchRelationDefinitions,new(chain)),
	M->>slot(sketchStructuralRelationDefinitions,new(chain)),
	M->>slot(sketches,new(chain)), % AB, feb 2006
        M->>initialise_sketch, % AB, feb 2006

	% Create the fields for the metadata (should already have been done in patch 6)
	send(M, createMissingMetaDataFields),

	(
	    get(@model, lastChangeTime, @nil) ->
	    new(LastChangeTime, date),
	    send(@model, slot, lastChangeTime, LastChangeTime)
	;
	    true
	).
	


pl_versionPatch(M,7):-
	%from 7 to 8 we added garpQuantity->quantityAssumptions
	
	@pce->>write_ln('7 --> 8: Added quantity assumptions'),
	M?quantities->>for_all(->>(@arg1,quantityAssumptions,create(chain))).
	
	
pl_versionPatch(M,6):-
	%from 6 to 7 we added garpModel->metadata
	
	@pce->>write_ln('6 --> 7: Added general possibility to save meta data in the model.'),
	M->>metaData(new(hash_table)).
%
pl_versionPatch(M,5):-
	%from 5 to 6 we added modelFragment<->active. Used to switch the use of modelfragment in the simulation on or off.
	
	@pce->>write_ln('5 --> 6: Added ability to switch certain model fragments off for simulation.'),
	M?modelFragments->>for_all(
		->>(@arg1,active,@on)).
%	
pl_versionPatch(M,4):-
	%from 4 to 5 we added mfViews: a hash used for saving views in de mfStructureEditor.
	
	@pce->>write_ln('4 --> 5: Added mfViews data to model: save views in model fragment structure editor.'),
	Views *= hash_table,
	%also add special entries @nil and 'Default view'
	Views->>append(@nil,new(hash_table)),
	Views->>append('Default view',new(hash_table)),
	M->>mfViews(Views).

	
pl_versionPatch(M,3):-
	%From 3 to 4 we added RunPrefs: a hash containing the 'algorithm assumptions'
	%of the engine, that are saved with the model.
	%These are exported to algorithm_option_switch/2 facts in the isa file.
	
	@pce->>write_ln('3 --> 4: Added RunPrefs data to model.'),
	M->>initRunPrefs.
	
pl_versionPatch(M,2):-
	%From 2 to 3 is just a bugfix. Somewhere between version 0 and 1 we set
	%the ?full value of correspondences to 'nil'?? Anyway: set them to @off
	%do the same for mirror

	M?modelFragments->>for_all(
		and(
			->>(?(@arg1,findElements,correspondence),for_all,
			if(
				or(@arg1?full == @nil,
					@arg1?full == nil),
				and(
					->>(@pce,write_ln,'2 --> 3 patch: Set correspondence in modelfragment', @arg1?fragment?name, 'explicitly to being a non-full correspondence'),
					->>(@arg1,slot,full,@off)
				)
			)),
			->>(?(@arg1,findElements,correspondence),for_all,
			if(
				or(@arg1?mirror == @nil,
					@arg1?mirror == nil),
				and(
					->>(@pce,write_ln,'2 --> 3 patch: Set correspondence in modelfragment', @arg1?fragment?name, 'explicitly to being a non-inverse correspondence'),
					->>(@arg1,slot,mirror,@off)
				)
			))
		)
	).	
	
pl_versionPatch(M,1):-
	%from 1 to 2: change the saved simulations to a double hash table
	%value is no longer just a chain, but also contains a hash table, with 
	%different entries (one of which is the chain)
	
	@pce->>write_ln('1 --> 2: change the saved simulations to make sure they can contain meta data (save version of used model)'),
	%modelversion is the posix value of the lastChangeTime
	New *= hash_table,
	HT *= var,
	M?savedSimulations->>for_all(
		and(
			assign(HT,create(hash_table)),
			->>(HT,append,modelVersion,number(0)), %for starters, out of date
			->>(HT,append,data,@arg2),
			->>(New,append,@arg1,HT))
		),
	M->>savedSimulations(New).

	
pl_versionPatch(M,0):-
	%al changes before implementation of this version patch are meant te be in version 0
	%check of @model?dqs nog deel uitmaakt van de definitieobjecten
	%dat mag namelijk niet meer sinds 2.0b6

	@pce->>write_ln('Patching version 0 model: model saved before garp3 0.3.2.'),
	%gp3: we now have a model version thing, so we can check better
	
	ignore(M?quantitySpaces->>delete(@model?dqs)),
	%en voeg @model?mzpqs toe
	if
		( \+ (M?quantitySpaces->>member(M?mzpqs)))
	then
	(
			@pce->>write_ln('*** Version Patch (2.0): added new MZP Quantity Space as not being internally the same as the derivative one'),
		%en als kopie van wel zichtbare versie
		M->>slot(mzpqs,quantitySpace(mzp,chain(
				valueReference('Plus',interval),
				valueReference('Zero',point),
				valueReference('Min',interval)),
				'Fixed quantity space for derivatives. If needed, you can use this one for quantities too.')),
	%en die voegen we gewoon toe aan de qs-en. Deze kan niet verwijderd worden
	M->>insertDefinitionObject(M?mzpqs)
	),
	
	%kijk welke quantityspace definities de dqs toestaan
	%dat wordt nu de mzp
	
	M?quantityDefinitions->>for_all(
		if(
			->>(@arg1,quantitySpaceAllowed,M?dqs),
			and(
				->>(@pce,write_ln,'*** Version Patch (2.0): Change allowed quantity spaces for', @arg1?name,' to accomodate changed mzp quantity space'),
				->>(@arg1,hyper,M?mzpqs,allowedQuantitySpace,quantityDefinition),
				->>(@arg1,delete_hypers,allowedQuantitySpace,@arg3 == M?dqs)
				)
		)
	),	
	
	%pas quantities aan: als de dqs als qs hebben, dan wordt het mpzqs
	M?quantities->>for_all(
		if(
			@arg1?quantitySpace == M?dqs,
			and(
				->>(@pce,write_ln,'*** Version Patch (2.0): Change quantity space for', @arg1?name,' in modelfragment', @arg1?fragment?name, 'to accomodate changed mzp quantity space'),
				->>(M, changeRequest, changeQuantity,@arg1?fragment, @nil, @arg1, @arg1?stateName,
					@arg1?definition, M?mzpqs, @arg1?remarks)
			)
		)
	),
	
	%en kopieer alle quantityspace-waarden als ze een element van dqs hebben, via een helper
	
	%gp3: removed the helper
	
	QS *= var,
	Huidige *= var,
	NewValues *= var,
	
	M?quantitySpaces->>for_all(
		and(
			assign(QS,@arg1),
			if(
				->>(QS?values,find,
					and(
						assign(Huidige,@arg1?id),
						->>(M?dqs?values,find, @arg1?id == Huidige)
						)
					
				),
				and(
					->>(@pce,write_ln,'*** Version Patch (2.0betas): Quantity Space', QS?name, 'contained beta 5 buggy values from old mzp quantity space copies. Fixed.'),
					assign(NewValues, ?(QS?values,map,@arg1?copyContent)),
					->>(M,changeRequest,changeQuantitySpace,QS,@nil,QS?name,NewValues,QS?remarks)
				)
			)
		)
	),
	
	%gp3 0.11: make sure the savedSimulations member of the model is a hash_table
	if
		@nil = @model<<-savedSimulations
	then
	(
		@pce->>write_ln('*** Garp3 0.11 patch: Creating the savedSimulations hash_table in @model'),
		@model->>savedSimulations(new(hash_table))
	),
	
	%gp3 0.1: Check all existing correspondences for set value of new vars @derivative and @mirror
	%0.11: also check all existing assumption instances for set value of new var instanceRoute
	%0.2.14 also check correspondences for full var (Garp3 0.3 work is to use it)
	M?modelFragments->>for_all(
		and(
			->>(?(@arg1,findElements,correspondence),for_all,
				and(
					if(
						@arg1?derivative == @nil,
						and(
							->>(@pce,write_ln,'*** Garp3 0.1 patch: Set correspondence in modelfragment', @arg1?fragment?name, 'explicitly to being a non-derivative correspondence'),
							->>(@arg1,slot,derivative,@off)
						)
					),
					if(
						@arg1?mirror == @nil,
						and(
							->>(@pce,write_ln,'*** Garp3 0.1 patch: Set correspondence in modelfragment', @arg1?fragment?name, 'explicitly to being a non-mirror correspondence'),
							->>(@arg1,slot,mirror,@off)
						)
					),
					if(
						or(@arg1?full == @nil,
							@arg1?full == nil),
						and(
							->>(@pce,write_ln,'*** Garp3 0.3 patch: Set correspondence in modelfragment', @arg1?fragment?name, 'explicitly to being a non-full correspondence'),
							->>(@arg1,slot,full,@off)
						)
					)
				)
			),
			->>(?(@arg1,findElements,assumptionInstance),for_all,
				if(
					?(@arg1,slot,instanceRoute) == @nil,
					and(
						->>(@pce,write_ln,'*** Garp3 0.11 patch: Set assumption instance in modelfragment', @arg1?fragment?name, 'to have an empty instanceRoute'),
						->>(@arg1,slot,instanceRoute,new(chain))
					)
				)
			)
		)
	),
	%gp3 0.11: check if the modelfragment names are correct
	unless
		'Static' = @model?topStaticFragment<<-name
	do
	(
		@pce->>write_ln(string('Garp3 0.1 patch: The name of the top static fragment has been changed from %s to Static',
			@model?topStaticFragment?name)),
		@model?topStaticFragment->>slot(name,'Static')
	),
	unless
		'Process' = @model?topProcessFragment<<-name
	do
	(
		@pce->>write_ln(string('Garp3 0.1 patch: The name of the top process fragment has been changed from %s to Process',
			@model?topProcessFragment?name)),
		@model?topProcessFragment->>slot(name,'Process')
	),
	unless
		'Agent' = @model?topAgentFragment<<-name
	do
	(
		@pce->>write_ln(string('Garp3 0.1 patch: The name of the top agent fragment has been changed from %s to Agent',
			@model?topAgentFragment?name)),
		@model?topAgentFragment->>slot(name,'Agent')
	),
	
	@pce->>write_ln('****\n\n').
%



%%%%%% some models somehow got a version number 9 or higher without this update

%
pl_checkMissingVersionPatch(M, 8):-
    % check for missing versionPatch for version 8   
    (
	% assuming that if this one is not present, or @nil, 
        % the update is necessary for all sketches
        get(M, slot, sketchScenarioDefinitions, ScenDefs),
        ScenDefs = @nil
    ->
	@pce->>write_ln('Missing update detected for version 8 --> 9.'),
        pl_versionPatch(M, 8),
        % and delete remnants of old (empty) sketch representations
        get(M?abstractEntities, find_all, message(@arg1, instance_of, sheet), OldSketchSheets), 
        unless
            OldSketchSheets->>empty
        do
        ( 
            send(OldSketchSheets, for_all, message(M?abstractEntities, delete, @arg1)),
            send(OldSketchSheets, for_all, message(@arg1, free))
        ),
	send(@model, modelDefinitionVersion, 9),
	send(@model, versionPatch)
    ;
	fail
    ).
%%


%%%%%% for updating Sketch definitions 

pl_updateSketchDefinitions(Defs):-
        send(Defs, for_all, 
              message(@prolog, pl_updateSketchDefinition, @arg1)
        ).

pl_updateSketchDefinition(Def):-
       send(Def?def_sheet?members, for_all, 
          message(@prolog, pl_updateSketchDefinitionAttribute, Def?def_sheet, @arg1)
       ).

pl_updateSketchDefinitionAttribute(DefSheet, Attr):-
       get(Attr, name, entities), !, 
       get(Attr?value, value, Str), 
       new(EntityChain, chain), 
       unless
		Str == ''
       do
        (
         swritef(RemarksStr, 'From old model version: %w',[Str]), 
         new(E, sketchDefinitionElement(Str, entity, RemarksStr)), 
         send(EntityChain, append, E)
        ),
       new(Attr2, attribute(entity, EntityChain)),
       send(DefSheet, delete, entities), 
       send(DefSheet, append, Attr2).

pl_updateSketchDefinitionAttribute(DefSheet, Attr):-
       get(Attr, name, quantities), !, 
       get(Attr?value, value, Str), 
       new(QuantityChain, chain), 
       unless
		Str == ''
       do
        (
         swritef(RemarksStr, 'From old model version: %w',[Str]), 
         new(Q, sketchDefinitionElement(Str, quantity, RemarksStr)), 
         send(QuantityChain, append, Q)
       ),
       new(Attr2, attribute(quantity, QuantityChain)),
       send(DefSheet, delete, quantities), 
       send(DefSheet, append, Attr2).

pl_updateSketchDefinitionAttribute(DefSheet, Attr):-
       get(Attr, name, agents), !, 
       get(Attr?value, value, Str), 
       new(AgentChain, chain), 
       unless
		Str == ''
       do
        (
         swritef(RemarksStr, 'From old model version: %w',[Str]), 
         new(A, sketchDefinitionElement(Str, agent, RemarksStr)), 
         send(AgentChain, append, A)
        ),
       new(Attr2, attribute(agent, AgentChain)),
       send(DefSheet, delete, agents), 
       send(DefSheet, append, Attr2).

pl_updateSketchDefinitionAttribute(DefSheet, Attr):-
       get(Attr, name, assumptions), !, 
       get(Attr?value, value, Str), 
       new(AssumptionChain, chain), 
       unless
		Str == ''
       do
        (
         swritef(RemarksStr, 'From old model version: %w',[Str]), 
         new(As, sketchDefinitionElement(Str, assumption, RemarksStr)), 
         send(AssumptionChain, append, As)
        ),
       new(Attr2, attribute(assumption, AssumptionChain)),
       send(DefSheet, delete, assumptions), 
       send(DefSheet, append, Attr2).

% rest case - no update necessary
pl_updateSketchDefinitionAttribute(_DefSheet, _Attr).
       

